"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getFeatureValueObject = exports.getFeatureMainValue = exports.getFeatureMainProperty = exports.processFeatureIdentifier = exports.getBackwardCompatibleFeatureId = exports.getFeatureIdType = exports.generateFeaturesConfig = exports.loadV1FeaturesJsonFromDisk = exports.loadFeaturesJson = exports.generateContainerEnvs = exports.getFeatureLayers = exports.getFeatureInstallWrapperScript = exports.getContainerFeaturesBaseDockerFile = exports.getSourceInfoString = exports.getContainerFeaturesFolder = exports.multiStageBuildExploration = exports.collapseFeaturesConfig = exports.parseMount = exports.DEVCONTAINER_FEATURE_FILE_NAME = exports.V1_DEVCONTAINER_FEATURES_FILE_NAME = void 0;
const jsonc = __importStar(require("jsonc-parser"));
const path = __importStar(require("path"));
const URL = __importStar(require("url"));
const tar = __importStar(require("tar"));
const pfs_1 = require("../spec-utils/pfs");
const log_1 = require("../spec-utils/log");
const httpRequest_1 = require("../spec-utils/httpRequest");
const containerFeaturesOrder_1 = require("./containerFeaturesOrder");
const containerFeaturesOCI_1 = require("./containerFeaturesOCI");
// v1
const V1_ASSET_NAME = 'devcontainer-features.tgz';
exports.V1_DEVCONTAINER_FEATURES_FILE_NAME = 'devcontainer-features.json';
// v2
exports.DEVCONTAINER_FEATURE_FILE_NAME = 'devcontainer-feature.json';
function parseMount(str) {
    return str.split(',')
        .map(s => s.split('='))
        .reduce((acc, [key, value]) => ({ ...acc, [key]: value }), {});
}
exports.parseMount = parseMount;
function collapseFeaturesConfig(original) {
    const collapsed = {
        allFeatures: original.featureSets
            .map(fSet => fSet.features)
            .flat()
    };
    return collapsed;
}
exports.collapseFeaturesConfig = collapseFeaturesConfig;
exports.multiStageBuildExploration = false;
// Counter to ensure that no two folders are the same even if we are executing the same feature multiple times.
let counter = 1;
function getCounter() {
    return counter++;
}
const isTsnode = path.basename(process.argv[0]) === 'ts-node' || process.argv.indexOf('ts-node/register') !== -1;
function getContainerFeaturesFolder(_extensionPath) {
    if (isTsnode) {
        return path.join(require.resolve('vscode-dev-containers/package.json'), '..', 'container-features');
    }
    const distFolder = typeof _extensionPath === 'string' ? path.join(_extensionPath, 'dist') : _extensionPath.distFolder;
    return path.join(distFolder, 'node_modules', 'vscode-dev-containers', 'container-features');
}
exports.getContainerFeaturesFolder = getContainerFeaturesFolder;
// Take a SourceInformation and condense it down into a single string
// Useful for calculating a unique build folder name for a given featureSet.
function getSourceInfoString(srcInfo) {
    const { type } = srcInfo;
    switch (type) {
        case 'local-cache':
            return 'local-cache-' + getCounter();
        case 'direct-tarball':
            return srcInfo.tarballUri + getCounter();
        case 'github-repo':
            return `github-${srcInfo.owner}-${srcInfo.repo}-${srcInfo.isLatest ? 'latest' : srcInfo.tag}-${getCounter()}`;
        case 'file-path':
            return srcInfo.resolvedFilePath + '-' + getCounter();
        case 'oci':
            return `oci-${srcInfo.featureRef.resource}-${getCounter()}`;
    }
}
exports.getSourceInfoString = getSourceInfoString;
// TODO: Move to node layer.
function getContainerFeaturesBaseDockerFile() {
    return `
#{featureBuildStages}

#{nonBuildKitFeatureContentFallback}

FROM $_DEV_CONTAINERS_BASE_IMAGE AS dev_containers_target_stage

USER root

COPY --from=dev_containers_feature_content_source {contentSourceRootPath} /tmp/build-features/

#{featureLayer}

#{copyFeatureBuildStages}

#{containerEnv}

ARG _DEV_CONTAINERS_IMAGE_USER=root
USER $_DEV_CONTAINERS_IMAGE_USER

#{devcontainerMetadata}
`;
}
exports.getContainerFeaturesBaseDockerFile = getContainerFeaturesBaseDockerFile;
function getFeatureInstallWrapperScript(feature, featureSet, options) {
    var _a, _b, _c, _d, _e;
    const id = escapeQuotesForShell((_a = featureSet.sourceInformation.userFeatureIdWithoutVersion) !== null && _a !== void 0 ? _a : 'Unknown');
    const name = escapeQuotesForShell((_b = feature.name) !== null && _b !== void 0 ? _b : 'Unknown');
    const description = escapeQuotesForShell((_c = feature.description) !== null && _c !== void 0 ? _c : '');
    const version = escapeQuotesForShell((_d = feature.version) !== null && _d !== void 0 ? _d : '');
    const documentation = escapeQuotesForShell((_e = feature.documentationURL) !== null && _e !== void 0 ? _e : '');
    const optionsIndented = escapeQuotesForShell(options.map(x => `    ${x}`).join('\n'));
    const errorMessage = `ERROR: Feature "${name}" (${id}) failed to install!`;
    const troubleshootingMessage = documentation
        ? ` Look at the documentation at ${documentation} for help troubleshooting this error.`
        : '';
    return `#!/bin/sh
set -e

on_exit () {
	[ $? -eq 0 ] && exit
	echo '${errorMessage}${troubleshootingMessage}'
}

trap on_exit EXIT

echo ===========================================================================
echo 'Feature       : ${name}'
echo 'Description   : ${description}'
echo 'Id            : ${id}'
echo 'Version       : ${version}'
echo 'Documentation : ${documentation}'
echo 'Options       :'
echo '${optionsIndented}'
echo ===========================================================================

set -a
. ./devcontainer-features.env
set +a

chmod +x ./install.sh
./install.sh
`;
}
exports.getFeatureInstallWrapperScript = getFeatureInstallWrapperScript;
function escapeQuotesForShell(input) {
    // The `input` is expected to be a string which will be printed inside single quotes
    // by the caller. This means we need to escape any nested single quotes within the string.
    // We can do this by ending the first string with a single quote ('), printing an escaped
    // single quote (\'), and then opening a new string (').
    return input.replace(new RegExp(`'`, 'g'), `'\\''`);
}
function getFeatureLayers(featuresConfig) {
    let result = '';
    // Features version 1
    const folders = (featuresConfig.featureSets || []).filter(y => y.internalVersion !== '2').map(x => x.features[0].consecutiveId);
    folders.forEach(folder => {
        result += `RUN cd /tmp/build-features/${folder} \\
&& chmod +x ./install.sh \\
&& ./install.sh

`;
    });
    // Features version 2
    featuresConfig.featureSets.filter(y => y.internalVersion === '2').forEach(featureSet => {
        featureSet.features.forEach(feature => {
            result += generateContainerEnvs(feature);
            result += `
RUN cd /tmp/build-features/${feature.consecutiveId} \\
&& chmod +x ./devcontainer-features-install.sh \\
&& ./devcontainer-features-install.sh

`;
        });
    });
    return result;
}
exports.getFeatureLayers = getFeatureLayers;
// Features version two export their environment variables as part of the Dockerfile to make them available to subsequent features.
function generateContainerEnvs(feature) {
    let result = '';
    if (!feature.containerEnv) {
        return result;
    }
    let keys = Object.keys(feature.containerEnv);
    result = keys.map(k => `ENV ${k}=${feature.containerEnv[k]}`).join('\n');
    return result;
}
exports.generateContainerEnvs = generateContainerEnvs;
const allowedFeatureIdRegex = new RegExp('^[a-zA-Z0-9_-]*$');
// Parses a declared feature in user's devcontainer file into
// a usable URI to download remote features.
// RETURNS
// {
//  "id",              <----- The ID of the feature in the feature set.
//  sourceInformation  <----- Source information (is this locally cached, a GitHub remote feature, etc..), including tarballUri if applicable.
// }
//
const cleanupIterationFetchAndMerge = async (tempTarballPath, output) => {
    // Non-fatal, will just get overwritten if we don't do the cleaned up.
    try {
        await (0, pfs_1.rmLocal)(tempTarballPath, { force: true });
    }
    catch (e) {
        output.write(`Didn't remove temporary tarball from disk with caught exception: ${e === null || e === void 0 ? void 0 : e.Message} `, log_1.LogLevel.Trace);
    }
};
function getRequestHeaders(sourceInformation, env, output) {
    let headers = {
        'user-agent': 'devcontainer'
    };
    const isGitHubUri = (srcInfo) => {
        const uri = srcInfo.tarballUri;
        return uri.startsWith('https://github.com') || uri.startsWith('https://api.github.com');
    };
    if (sourceInformation.type === 'github-repo' || (sourceInformation.type === 'direct-tarball' && isGitHubUri(sourceInformation))) {
        const githubToken = env['GITHUB_TOKEN'];
        if (githubToken) {
            output.write('Using environment GITHUB_TOKEN.');
            headers.Authorization = `Bearer ${githubToken}`;
        }
        else {
            output.write('No environment GITHUB_TOKEN available.');
        }
    }
    return headers;
}
async function askGitHubApiForTarballUri(sourceInformation, feature, headers, output) {
    const options = {
        type: 'GET',
        url: sourceInformation.apiUri,
        headers
    };
    const apiInfo = JSON.parse(((await (0, httpRequest_1.request)(options, output)).toString()));
    if (apiInfo) {
        const asset = apiInfo.assets.find(a => a.name === `${feature.id}.tgz`) // v2
            || apiInfo.assets.find(a => a.name === V1_ASSET_NAME) // v1
            || undefined;
        if (asset && asset.url) {
            output.write(`Found url to fetch release artifact '${asset.name}'. Asset of size ${asset.size} has been downloaded ${asset.download_count} times and was last updated at ${asset.updated_at}`);
            return asset.url;
        }
        else {
            output.write('Unable to fetch release artifact URI from GitHub API', log_1.LogLevel.Error);
            return undefined;
        }
    }
    return undefined;
}
async function loadFeaturesJson(jsonBuffer, filePath, output) {
    if (jsonBuffer.length === 0) {
        output.write('Parsed featureSet is empty.', log_1.LogLevel.Error);
        return undefined;
    }
    const featureSet = jsonc.parse(jsonBuffer.toString());
    if (!(featureSet === null || featureSet === void 0 ? void 0 : featureSet.features) || featureSet.features.length === 0) {
        output.write('Parsed featureSet contains no features.', log_1.LogLevel.Error);
        return undefined;
    }
    output.write(`Loaded ${filePath}, which declares ${featureSet.features.length} features and ${(!!featureSet.sourceInformation) ? 'contains' : 'does not contain'} explicit source info.`, log_1.LogLevel.Trace);
    return updateFromOldProperties(featureSet);
}
exports.loadFeaturesJson = loadFeaturesJson;
async function loadV1FeaturesJsonFromDisk(pathToDirectory, output) {
    const filePath = path.join(pathToDirectory, exports.V1_DEVCONTAINER_FEATURES_FILE_NAME);
    const jsonBuffer = await (0, pfs_1.readLocalFile)(filePath);
    return loadFeaturesJson(jsonBuffer, filePath, output);
}
exports.loadV1FeaturesJsonFromDisk = loadV1FeaturesJsonFromDisk;
function updateFromOldProperties(original) {
    // https://github.com/microsoft/dev-container-spec/issues/1
    if (!original.features.find(f => f.extensions || f.settings)) {
        return original;
    }
    return {
        ...original,
        features: original.features.map(f => {
            if (!(f.extensions || f.settings)) {
                return f;
            }
            const copy = { ...f };
            const customizations = copy.customizations || (copy.customizations = {});
            const vscode = customizations.vscode || (customizations.vscode = {});
            if (copy.extensions) {
                vscode.extensions = (vscode.extensions || []).concat(copy.extensions);
                delete copy.extensions;
            }
            if (copy.settings) {
                vscode.settings = {
                    ...copy.settings,
                    ...(vscode.settings || {}),
                };
                delete copy.settings;
            }
            return copy;
        }),
    };
}
// Generate a base featuresConfig object with the set of locally-cached features, 
// as well as downloading and merging in remote feature definitions.
async function generateFeaturesConfig(params, dstFolder, config, getLocalFeaturesFolder) {
    const { output } = params;
    const workspaceRoot = params.cwd;
    output.write(`workspace root: ${workspaceRoot}`, log_1.LogLevel.Trace);
    const userFeatures = featuresToArray(config);
    if (!userFeatures) {
        return undefined;
    }
    // Create the featuresConfig object.
    // Initialize the featureSets object, and stash the dstFolder on the object for use later.
    let featuresConfig = {
        featureSets: [],
        dstFolder
    };
    // load local cache of features;
    // TODO: Update so that cached features are always version 2
    const localFeaturesFolder = getLocalFeaturesFolder(params.extensionPath);
    const locallyCachedFeatureSet = await loadV1FeaturesJsonFromDisk(localFeaturesFolder, output); // TODO: Pass dist folder instead to also work with the devcontainer.json support package.
    if (!locallyCachedFeatureSet) {
        output.write('Failed to load locally cached features', log_1.LogLevel.Error);
        return undefined;
    }
    // Read features and get the type.
    output.write('--- Processing User Features ----', log_1.LogLevel.Trace);
    featuresConfig = await processUserFeatures(params.output, config, workspaceRoot, params.env, userFeatures, featuresConfig, params.skipFeatureAutoMapping);
    output.write(JSON.stringify(featuresConfig, null, 4), log_1.LogLevel.Trace);
    const ociCacheDir = await prepareOCICache(dstFolder);
    // Fetch features, stage into the appropriate build folder, and read the feature's devcontainer-feature.json
    output.write('--- Fetching User Features ----', log_1.LogLevel.Trace);
    await fetchFeatures(params, featuresConfig, locallyCachedFeatureSet, dstFolder, localFeaturesFolder, ociCacheDir);
    const orderedFeatures = (0, containerFeaturesOrder_1.computeFeatureInstallationOrder)(config, featuresConfig.featureSets);
    output.write('--- Computed order ----', log_1.LogLevel.Trace);
    for (const feature of orderedFeatures) {
        output.write(`${feature.sourceInformation.userFeatureId}`, log_1.LogLevel.Trace);
    }
    featuresConfig.featureSets = orderedFeatures;
    return featuresConfig;
}
exports.generateFeaturesConfig = generateFeaturesConfig;
async function prepareOCICache(dstFolder) {
    const ociCacheDir = path.join(dstFolder, 'ociCache');
    await (0, pfs_1.mkdirpLocal)(ociCacheDir);
    return ociCacheDir;
}
function featuresToArray(config) {
    if (!config.features) {
        return undefined;
    }
    const userFeatures = [];
    for (const userFeatureKey of Object.keys(config.features)) {
        const userFeatureValue = config.features[userFeatureKey];
        const feature = {
            id: userFeatureKey,
            options: userFeatureValue
        };
        userFeatures.push(feature);
    }
    return userFeatures;
}
// Process features contained in devcontainer.json
// Creates one feature set per feature to aid in support of the previous structure.
async function processUserFeatures(output, config, workspaceRoot, env, userFeatures, featuresConfig, skipFeatureAutoMapping) {
    for (const userFeature of userFeatures) {
        const newFeatureSet = await processFeatureIdentifier(output, config.configFilePath.path, workspaceRoot, env, userFeature, skipFeatureAutoMapping);
        if (!newFeatureSet) {
            throw new Error(`Failed to process feature ${userFeature.id}`);
        }
        featuresConfig.featureSets.push(newFeatureSet);
    }
    return featuresConfig;
}
async function getFeatureIdType(output, env, userFeatureId) {
    // See the specification for valid feature identifiers:
    //   > https://github.com/devcontainers/spec/blob/main/proposals/devcontainer-features.md#referencing-a-feature
    //
    // Additionally, we support the following deprecated syntaxes for backwards compatibility:
    //      (0)  A 'local feature' packaged with the CLI.
    //			 Syntax:   <feature>
    //
    //      (1)  A feature backed by a GitHub Release
    //			 Syntax:   <repoOwner>/<repoName>/<featureId>[@version]
    // DEPRECATED: This is a legacy feature-set ID
    if (!userFeatureId.includes('/') && !userFeatureId.includes('\\')) {
        return { type: 'local-cache', manifest: undefined };
    }
    // Direct tarball reference
    if (userFeatureId.startsWith('https://')) {
        return { type: 'direct-tarball', manifest: undefined };
    }
    // Local feature on disk
    // !! NOTE: The ability for paths outside the project file tree will soon be removed.
    if (userFeatureId.startsWith('./') || userFeatureId.startsWith('../') || userFeatureId.startsWith('/')) {
        return { type: 'file-path', manifest: undefined };
    }
    // version identifier for a github release feature.
    // DEPRECATED: This is a legacy feature-set ID
    if (userFeatureId.includes('@')) {
        return { type: 'github-repo', manifest: undefined };
    }
    const manifest = await (0, containerFeaturesOCI_1.fetchOCIFeatureManifestIfExistsFromUserIdentifier)(output, env, userFeatureId);
    if (manifest) {
        return { type: 'oci', manifest: manifest };
    }
    else {
        // DEPRECATED: This is a legacy feature-set ID
        output.write('(!) WARNING: Falling back to deprecated GitHub Release syntax. See https://github.com/devcontainers/spec/blob/main/proposals/devcontainer-features.md#referencing-a-feature for updated specification.', log_1.LogLevel.Warning);
        return { type: 'github-repo', manifest: undefined };
    }
}
exports.getFeatureIdType = getFeatureIdType;
function getBackwardCompatibleFeatureId(output, id) {
    const migratedfeatures = ['aws-cli', 'azure-cli', 'desktop-lite', 'docker-in-docker', 'docker-from-docker', 'dotnet', 'git', 'git-lfs', 'github-cli', 'java', 'kubectl-helm-minikube', 'node', 'powershell', 'python', 'ruby', 'rust', 'sshd', 'terraform'];
    const renamedFeatures = new Map();
    renamedFeatures.set('golang', 'go');
    renamedFeatures.set('common', 'common-utils');
    const deprecatedFeaturesIntoOptions = new Map();
    deprecatedFeaturesIntoOptions.set('gradle', 'java');
    deprecatedFeaturesIntoOptions.set('maven', 'java');
    deprecatedFeaturesIntoOptions.set('jupyterlab', 'python');
    // TODO: add warning logs once we have context on the new location for these Features.
    // const deprecatedFeatures = ['fish', 'homebrew'];
    const newFeaturePath = 'ghcr.io/devcontainers/features';
    // Note: Pin the versionBackwardComp to '1' to avoid breaking changes.
    const versionBackwardComp = '1';
    // Mapping feature references (old shorthand syntax) from "microsoft/vscode-dev-containers" to "ghcr.io/devcontainers/features"
    if (migratedfeatures.includes(id)) {
        output.write(`(!) WARNING: Using the deprecated '${id}' Feature. See https://github.com/devcontainers/features/tree/main/src/${id}#example-usage for the updated Feature.`, log_1.LogLevel.Warning);
        return `${newFeaturePath}/${id}:${versionBackwardComp}`;
    }
    // Mapping feature references (renamed old shorthand syntax) from "microsoft/vscode-dev-containers" to "ghcr.io/devcontainers/features"
    if (renamedFeatures.get(id) !== undefined) {
        output.write(`(!) WARNING: Using the deprecated '${id}' Feature. See https://github.com/devcontainers/features/tree/main/src/${renamedFeatures.get(id)}#example-usage for the updated Feature.`, log_1.LogLevel.Warning);
        return `${newFeaturePath}/${renamedFeatures.get(id)}:${versionBackwardComp}`;
    }
    if (deprecatedFeaturesIntoOptions.get(id) !== undefined) {
        output.write(`(!) WARNING: Falling back to the deprecated '${id}' Feature. It is now part of the '${deprecatedFeaturesIntoOptions.get(id)}' Feature. See https://github.com/devcontainers/features/tree/main/src/${deprecatedFeaturesIntoOptions.get(id)}#options for the updated Feature.`, log_1.LogLevel.Warning);
    }
    // Deprecated and all other features references (eg. fish, ghcr.io/devcontainers/features/go, ghcr.io/owner/repo/id etc)
    return id;
}
exports.getBackwardCompatibleFeatureId = getBackwardCompatibleFeatureId;
// Strictly processes the user provided feature identifier to determine sourceInformation type.
// Returns a featureSet per feature.
async function processFeatureIdentifier(output, configPath, _workspaceRoot, env, userFeature, skipFeatureAutoMapping) {
    output.write(`* Processing feature: ${userFeature.id}`);
    // id referenced by the user before the automapping from old shorthand syntax to "ghcr.io/devcontainers/features"
    const originalUserFeatureId = userFeature.id;
    // Adding backward compatibility
    if (!skipFeatureAutoMapping) {
        userFeature.id = getBackwardCompatibleFeatureId(output, userFeature.id);
    }
    const { type, manifest } = await getFeatureIdType(output, env, userFeature.id);
    // cached feature
    // Resolves deprecated features (fish, maven, gradle, homebrew, jupyterlab)
    if (type === 'local-cache') {
        output.write(`Cached feature found.`);
        let feat = {
            id: userFeature.id,
            name: userFeature.id,
            value: userFeature.options,
            included: true,
        };
        let newFeaturesSet = {
            sourceInformation: {
                type: 'local-cache',
                userFeatureId: originalUserFeatureId
            },
            features: [feat],
        };
        return newFeaturesSet;
    }
    // remote tar file
    if (type === 'direct-tarball') {
        output.write(`Remote tar file found.`);
        const tarballUri = new URL.URL(userFeature.id);
        const fullPath = tarballUri.pathname;
        const tarballName = fullPath.substring(fullPath.lastIndexOf('/') + 1);
        output.write(`tarballName = ${tarballName}`, log_1.LogLevel.Trace);
        const regex = new RegExp('devcontainer-feature-(.*).tgz');
        const matches = regex.exec(tarballName);
        if (!matches || matches.length !== 2) {
            output.write(`Expected tarball name to follow 'devcontainer-feature-<feature-id>.tgz' format.  Received '${tarballName}'`, log_1.LogLevel.Error);
            return undefined;
        }
        const id = matches[1];
        if (id === '' || !allowedFeatureIdRegex.test(id)) {
            output.write(`Parse error. Specify a feature id with alphanumeric, dash, or underscore characters. Received ${id}.`, log_1.LogLevel.Error);
            return undefined;
        }
        let feat = {
            id: id,
            name: userFeature.id,
            value: userFeature.options,
            included: true,
        };
        let newFeaturesSet = {
            sourceInformation: {
                type: 'direct-tarball',
                tarballUri: tarballUri.toString(),
                userFeatureId: originalUserFeatureId
            },
            features: [feat],
        };
        return newFeaturesSet;
    }
    // Spec: https://containers.dev/implementors/features-distribution/#addendum-locally-referenced
    if (type === 'file-path') {
        output.write(`Local disk feature.`);
        const id = path.basename(userFeature.id);
        // Fail on Absolute paths.
        if (path.isAbsolute(userFeature.id)) {
            output.write('An Absolute path to a local feature is not allowed.', log_1.LogLevel.Error);
            return undefined;
        }
        // Local-path features are expected to be a sub-folder of the '$WORKSPACE_ROOT/.devcontainer' folder.
        const featureFolderPath = path.join(path.dirname(configPath), userFeature.id);
        // Ensure we aren't escaping .devcontainer folder
        const parent = path.join(_workspaceRoot, '.devcontainer');
        const child = featureFolderPath;
        const relative = path.relative(parent, child);
        output.write(`${parent} -> ${child}:   Relative Distance = '${relative}'`, log_1.LogLevel.Trace);
        if (relative.indexOf('..') !== -1) {
            output.write(`Local file path parse error. Resolved path must be a child of the .devcontainer/ folder.  Parsed: ${featureFolderPath}`, log_1.LogLevel.Error);
            return undefined;
        }
        output.write(`Resolved: ${userFeature.id}  ->  ${featureFolderPath}`, log_1.LogLevel.Trace);
        // -- All parsing and validation steps complete at this point.
        output.write(`Parsed feature id: ${id}`, log_1.LogLevel.Trace);
        let feat = {
            id,
            name: userFeature.id,
            value: userFeature.options,
            included: true,
        };
        let newFeaturesSet = {
            sourceInformation: {
                type: 'file-path',
                resolvedFilePath: featureFolderPath,
                userFeatureId: originalUserFeatureId
            },
            features: [feat],
        };
        return newFeaturesSet;
    }
    // (6) Oci Identifier
    if (type === 'oci' && manifest) {
        let newFeaturesSet = (0, containerFeaturesOCI_1.getOCIFeatureSet)(output, userFeature.id, userFeature.options, manifest, originalUserFeatureId);
        return newFeaturesSet;
    }
    output.write(`Github feature.`);
    // Github repository source.
    let version = 'latest';
    let splitOnAt = userFeature.id.split('@');
    if (splitOnAt.length > 2) {
        output.write(`Parse error. Use the '@' symbol only to designate a version tag.`, log_1.LogLevel.Error);
        return undefined;
    }
    if (splitOnAt.length === 2) {
        output.write(`[${userFeature.id}] has version ${splitOnAt[1]}`, log_1.LogLevel.Trace);
        version = splitOnAt[1];
    }
    // Remaining info must be in the first part of the split.
    const featureBlob = splitOnAt[0];
    const splitOnSlash = featureBlob.split('/');
    // We expect all GitHub/registry features to follow the triple slash pattern at this point
    //  eg: <publisher>/<feature-set>/<feature>
    if (splitOnSlash.length !== 3 || splitOnSlash.some(x => x === '') || !allowedFeatureIdRegex.test(splitOnSlash[2])) {
        output.write(`Invalid parse for GitHub Release feature: Follow format '<publisher>/<feature-set>/<feature>, or republish feature to OCI registry.'`, log_1.LogLevel.Error);
        return undefined;
    }
    const owner = splitOnSlash[0];
    const repo = splitOnSlash[1];
    const id = splitOnSlash[2];
    let feat = {
        id: id,
        name: userFeature.id,
        value: userFeature.options,
        included: true,
    };
    const userFeatureIdWithoutVersion = originalUserFeatureId.split('@')[0];
    if (version === 'latest') {
        let newFeaturesSet = {
            sourceInformation: {
                type: 'github-repo',
                apiUri: `https://api.github.com/repos/${owner}/${repo}/releases/latest`,
                unauthenticatedUri: `https://github.com/${owner}/${repo}/releases/latest/download`,
                owner,
                repo,
                isLatest: true,
                userFeatureId: originalUserFeatureId,
                userFeatureIdWithoutVersion
            },
            features: [feat],
        };
        return newFeaturesSet;
    }
    else {
        // We must have a tag, return a tarball URI for the tagged version. 
        let newFeaturesSet = {
            sourceInformation: {
                type: 'github-repo',
                apiUri: `https://api.github.com/repos/${owner}/${repo}/releases/tags/${version}`,
                unauthenticatedUri: `https://github.com/${owner}/${repo}/releases/download/${version}`,
                owner,
                repo,
                tag: version,
                isLatest: false,
                userFeatureId: originalUserFeatureId,
                userFeatureIdWithoutVersion
            },
            features: [feat],
        };
        return newFeaturesSet;
    }
    // TODO: Handle invalid source types better by refactoring this function.
    // throw new Error(`Unsupported feature source type: ${type}`);
}
exports.processFeatureIdentifier = processFeatureIdentifier;
async function fetchFeatures(params, featuresConfig, localFeatures, dstFolder, localFeaturesFolder, ociCacheDir) {
    var _a, _b;
    for (const featureSet of featuresConfig.featureSets) {
        try {
            if (!featureSet || !featureSet.features || !featureSet.sourceInformation) {
                continue;
            }
            if (!localFeatures) {
                continue;
            }
            const { output } = params;
            const feature = featureSet.features[0];
            const consecutiveId = feature.id + '_' + getCounter();
            // Calculate some predictable caching paths.
            const featCachePath = path.join(dstFolder, consecutiveId);
            const sourceInfoType = (_a = featureSet.sourceInformation) === null || _a === void 0 ? void 0 : _a.type;
            feature.cachePath = featCachePath;
            feature.consecutiveId = consecutiveId;
            const featureDebugId = `${feature.consecutiveId}_${sourceInfoType}`;
            output.write(`* Fetching feature: ${featureDebugId}`);
            if (sourceInfoType === 'oci') {
                output.write(`Fetching from OCI`, log_1.LogLevel.Trace);
                await (0, pfs_1.mkdirpLocal)(featCachePath);
                const success = await (0, containerFeaturesOCI_1.fetchOCIFeature)(output, params.env, featureSet, ociCacheDir, featCachePath);
                if (!success) {
                    const err = `Could not download OCI feature: ${featureSet.sourceInformation.featureRef.id}`;
                    throw new Error(err);
                }
                if (!(await applyFeatureConfigToFeature(output, featureSet, feature, featCachePath))) {
                    const err = `Failed to parse feature '${featureDebugId}'. Please check your devcontainer.json 'features' attribute.`;
                    throw new Error(err);
                }
                continue;
            }
            if (sourceInfoType === 'local-cache') {
                // create copy of the local features to set the environment variables for them.
                await (0, pfs_1.mkdirpLocal)(featCachePath);
                await (0, pfs_1.cpDirectoryLocal)(localFeaturesFolder, featCachePath);
                if (!(await applyFeatureConfigToFeature(output, featureSet, feature, featCachePath))) {
                    const err = `Failed to parse feature '${featureDebugId}'. Please check your devcontainer.json 'features' attribute.`;
                    throw new Error(err);
                }
                continue;
            }
            if (sourceInfoType === 'file-path') {
                output.write(`Detected local file path`, log_1.LogLevel.Trace);
                await (0, pfs_1.mkdirpLocal)(featCachePath);
                const executionPath = featureSet.sourceInformation.resolvedFilePath;
                await (0, pfs_1.cpDirectoryLocal)(executionPath, featCachePath);
                if (!(await applyFeatureConfigToFeature(output, featureSet, feature, featCachePath))) {
                    const err = `Failed to parse feature '${featureDebugId}'. Please check your devcontainer.json 'features' attribute.`;
                    throw new Error(err);
                }
                continue;
            }
            output.write(`Detected tarball`, log_1.LogLevel.Trace);
            const headers = getRequestHeaders(featureSet.sourceInformation, params.env, output);
            // Ordered list of tarballUris to attempt to fetch from.
            let tarballUris = [];
            if (sourceInfoType === 'github-repo') {
                output.write('Determining tarball URI for provided github repo.', log_1.LogLevel.Trace);
                if (headers.Authorization && headers.Authorization !== '') {
                    output.write('GITHUB_TOKEN available. Attempting to fetch via GH API.', log_1.LogLevel.Info);
                    const authenticatedGithubTarballUri = await askGitHubApiForTarballUri(featureSet.sourceInformation, feature, headers, output);
                    if (authenticatedGithubTarballUri) {
                        tarballUris.push(authenticatedGithubTarballUri);
                    }
                    else {
                        output.write('Failed to generate autenticated tarball URI for provided feature, despite a GitHub token present', log_1.LogLevel.Warning);
                    }
                    headers.Accept = 'Accept: application/octet-stream';
                }
                // Always add the unauthenticated URIs as fallback options.
                output.write('Appending unauthenticated URIs for v2 and then v1', log_1.LogLevel.Trace);
                tarballUris.push(`${featureSet.sourceInformation.unauthenticatedUri}/${feature.id}.tgz`);
                tarballUris.push(`${featureSet.sourceInformation.unauthenticatedUri}/${V1_ASSET_NAME}`);
            }
            else {
                // We have a plain ol' tarball URI, since we aren't in the github-repo case.
                tarballUris.push(featureSet.sourceInformation.tarballUri);
            }
            // Attempt to fetch from 'tarballUris' in order, until one succeeds.
            let didSucceed = false;
            for (const tarballUri of tarballUris) {
                didSucceed = await fetchContentsAtTarballUri(tarballUri, featCachePath, headers, dstFolder, output);
                if (didSucceed) {
                    output.write(`Succeeded fetching ${tarballUri}`, log_1.LogLevel.Trace);
                    if (!(await applyFeatureConfigToFeature(output, featureSet, feature, featCachePath))) {
                        const err = `Failed to parse feature '${featureDebugId}'. Please check your devcontainer.json 'features' attribute.`;
                        throw new Error(err);
                    }
                    break;
                }
            }
            if (!didSucceed) {
                const msg = `(!) Failed to fetch tarball for ${featureDebugId} after attempting ${tarballUris.length} possibilities.`;
                throw new Error(msg);
            }
        }
        catch (e) {
            params.output.write(`(!) ERR: Failed to fetch feature: ${(_b = e === null || e === void 0 ? void 0 : e.message) !== null && _b !== void 0 ? _b : ''} `, log_1.LogLevel.Error);
            throw e;
        }
    }
}
async function fetchContentsAtTarballUri(tarballUri, featCachePath, headers, dstFolder, output) {
    const tempTarballPath = path.join(dstFolder, 'temp.tgz');
    try {
        const options = {
            type: 'GET',
            url: tarballUri,
            headers
        };
        output.write(`Fetching tarball at ${options.url}`);
        output.write(`Headers: ${JSON.stringify(options)}`, log_1.LogLevel.Trace);
        const tarball = await (0, httpRequest_1.request)(options, output);
        if (!tarball || tarball.length === 0) {
            output.write(`Did not receive a response from tarball download URI: ${tarballUri}`, log_1.LogLevel.Trace);
            return false;
        }
        // Filter what gets emitted from the tar.extract().
        const filter = (file, _) => {
            // Don't include .dotfiles or the archive itself.
            if (file.startsWith('./.') || file === `./${V1_ASSET_NAME}` || file === './.') {
                return false;
            }
            return true;
        };
        output.write(`Preparing to unarchive received tgz from ${tempTarballPath} -> ${featCachePath}.`, log_1.LogLevel.Trace);
        // Create the directory to cache this feature-set in.
        await (0, pfs_1.mkdirpLocal)(featCachePath);
        await (0, pfs_1.writeLocalFile)(tempTarballPath, tarball);
        await tar.x({
            file: tempTarballPath,
            cwd: featCachePath,
            filter
        });
        await cleanupIterationFetchAndMerge(tempTarballPath, output);
        return true;
    }
    catch (e) {
        output.write(`Caught failure when fetching from URI '${tarballUri}': ${e}`, log_1.LogLevel.Trace);
        await cleanupIterationFetchAndMerge(tempTarballPath, output);
        return false;
    }
}
// Reads the feature's 'devcontainer-feature.json` and applies any attributes to the in-memory Feature object.
// NOTE:
// 		Implements the latest ('internalVersion' = '2') parsing logic, 
// 		Falls back to earlier implementation(s) if requirements not present.
// 		Returns a boolean indicating whether the feature was successfully parsed.
async function applyFeatureConfigToFeature(output, featureSet, feature, featCachePath) {
    const innerJsonPath = path.join(featCachePath, exports.DEVCONTAINER_FEATURE_FILE_NAME);
    if (!(await (0, pfs_1.isLocalFile)(innerJsonPath))) {
        output.write(`Feature ${feature.id} is not a 'v2' feature. Attempting fallback to 'v1' implementation.`, log_1.LogLevel.Trace);
        output.write(`For v2, expected devcontainer-feature.json at ${innerJsonPath}`, log_1.LogLevel.Trace);
        return await parseDevContainerFeature_v1Impl(output, featureSet, feature, featCachePath);
    }
    featureSet.internalVersion = '2';
    feature.cachePath = featCachePath;
    const jsonString = await (0, pfs_1.readLocalFile)(innerJsonPath);
    const featureJson = jsonc.parse(jsonString.toString());
    feature = {
        ...featureJson,
        ...feature
    };
    featureSet.features[0] = updateFromOldProperties({ features: [feature] }).features[0];
    return true;
}
async function parseDevContainerFeature_v1Impl(output, featureSet, feature, featCachePath) {
    const pathToV1DevContainerFeatureJson = path.join(featCachePath, exports.V1_DEVCONTAINER_FEATURES_FILE_NAME);
    if (!(await (0, pfs_1.isLocalFile)(pathToV1DevContainerFeatureJson))) {
        output.write(`Failed to find ${exports.V1_DEVCONTAINER_FEATURES_FILE_NAME} metadata file (v1)`, log_1.LogLevel.Error);
        return false;
    }
    featureSet.internalVersion = '1';
    feature.cachePath = featCachePath;
    const jsonString = await (0, pfs_1.readLocalFile)(pathToV1DevContainerFeatureJson);
    const featureJson = jsonc.parse(jsonString.toString());
    const seekedFeature = featureJson === null || featureJson === void 0 ? void 0 : featureJson.features.find(f => f.id === feature.id);
    if (!seekedFeature) {
        output.write(`Failed to find feature '${feature.id}' in provided v1 metadata file`, log_1.LogLevel.Error);
        return false;
    }
    feature = {
        ...seekedFeature,
        ...feature
    };
    featureSet.features[0] = updateFromOldProperties({ features: [feature] }).features[0];
    return true;
}
function getFeatureMainProperty(feature) {
    var _a;
    return ((_a = feature.options) === null || _a === void 0 ? void 0 : _a.version) ? 'version' : undefined;
}
exports.getFeatureMainProperty = getFeatureMainProperty;
function getFeatureMainValue(feature) {
    var _a, _b;
    const defaultProperty = getFeatureMainProperty(feature);
    if (!defaultProperty) {
        return !!feature.value;
    }
    if (typeof feature.value === 'object') {
        const value = feature.value[defaultProperty];
        if (value === undefined && feature.options) {
            return (_a = feature.options[defaultProperty]) === null || _a === void 0 ? void 0 : _a.default;
        }
        return value;
    }
    if (feature.value === undefined && feature.options) {
        return (_b = feature.options[defaultProperty]) === null || _b === void 0 ? void 0 : _b.default;
    }
    return feature.value;
}
exports.getFeatureMainValue = getFeatureMainValue;
function getFeatureValueObject(feature) {
    if (typeof feature.value === 'object') {
        return {
            ...getFeatureValueDefaults(feature),
            ...feature.value
        };
    }
    const mainProperty = getFeatureMainProperty(feature);
    if (!mainProperty) {
        return getFeatureValueDefaults(feature);
    }
    return {
        ...getFeatureValueDefaults(feature),
        [mainProperty]: feature.value,
    };
}
exports.getFeatureValueObject = getFeatureValueObject;
function getFeatureValueDefaults(feature) {
    const options = feature.options || {};
    return Object.keys(options)
        .reduce((defaults, key) => {
        if ('default' in options[key]) {
            defaults[key] = options[key].default;
        }
        return defaults;
    }, {});
}
//# sourceMappingURL=containerFeaturesConfiguration.js.map