/// <reference types="node" />
import { Log } from '../spec-utils/log';
import { FeatureSet } from './containerFeaturesConfiguration';
export declare type HEADERS = {
    'authorization'?: string;
    'user-agent': string;
    'content-type'?: string;
    'accept'?: string;
};
export declare const DEVCONTAINER_MANIFEST_MEDIATYPE = "application/vnd.devcontainers";
export declare const DEVCONTAINER_TAR_LAYER_MEDIATYPE = "application/vnd.devcontainers.layer.v1+tar";
export declare const DEVCONTAINER_COLLECTION_LAYER_MEDIATYPE = "application/vnd.devcontainers.collection.layer.v1+json";
export interface OCIFeatureRef {
    registry: string;
    owner: string;
    namespace: string;
    path: string;
    resource: string;
    id: string;
    version?: string;
}
export interface OCIFeatureCollectionRef {
    registry: string;
    path: string;
    version: 'latest';
}
export interface OCILayer {
    mediaType: string;
    digest: string;
    size: number;
    annotations: {
        'org.opencontainers.image.title': string;
    };
}
export interface OCIManifest {
    digest?: string;
    schemaVersion: number;
    mediaType: string;
    config: {
        digest: string;
        mediaType: string;
        size: number;
    };
    layers: OCILayer[];
    annotations?: {};
}
export declare function getOCIFeatureSet(output: Log, identifier: string, options: boolean | string | Record<string, boolean | string | undefined>, manifest: OCIManifest, originalUserFeatureId: string): FeatureSet;
export declare function getFeatureRef(output: Log, resourceAndVersion: string): OCIFeatureRef;
export declare function fetchOCIFeatureManifestIfExistsFromUserIdentifier(output: Log, env: NodeJS.ProcessEnv, identifier: string, manifestDigest?: string, authToken?: string): Promise<OCIManifest | undefined>;
export declare function fetchOCIFeatureManifestIfExists(output: Log, env: NodeJS.ProcessEnv, featureRef: OCIFeatureRef | OCIFeatureCollectionRef, manifestDigest?: string, authToken?: string): Promise<OCIManifest | undefined>;
export declare function fetchOCIFeature(output: Log, env: NodeJS.ProcessEnv, featureSet: FeatureSet, ociCacheDir: string, featCachePath: string): Promise<boolean>;
export declare function getFeatureManifest(output: Log, env: NodeJS.ProcessEnv, url: string, featureRef: OCIFeatureRef | OCIFeatureCollectionRef, authToken?: string): Promise<OCIManifest | undefined>;
export declare function getFeatureBlob(output: Log, env: NodeJS.ProcessEnv, url: string, ociCacheDir: string, featCachePath: string, featureRef: OCIFeatureRef, authToken?: string): Promise<boolean>;
export declare function fetchRegistryAuthToken(output: Log, registry: string, ociRepoPath: string, env: NodeJS.ProcessEnv, operationScopes: string): Promise<string | undefined>;
export declare function getPublishedVersions(featureRef: OCIFeatureRef, output: Log, sorted?: boolean): Promise<string[] | undefined>;
