"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getPublishedVersions = exports.fetchRegistryAuthToken = exports.getFeatureBlob = exports.getFeatureManifest = exports.fetchOCIFeature = exports.fetchOCIFeatureManifestIfExists = exports.fetchOCIFeatureManifestIfExistsFromUserIdentifier = exports.getFeatureRef = exports.getOCIFeatureSet = exports.DEVCONTAINER_COLLECTION_LAYER_MEDIATYPE = exports.DEVCONTAINER_TAR_LAYER_MEDIATYPE = exports.DEVCONTAINER_MANIFEST_MEDIATYPE = void 0;
const path = __importStar(require("path"));
const tar = __importStar(require("tar"));
const semver = __importStar(require("semver"));
const httpRequest_1 = require("../spec-utils/httpRequest");
const log_1 = require("../spec-utils/log");
const pfs_1 = require("../spec-utils/pfs");
exports.DEVCONTAINER_MANIFEST_MEDIATYPE = 'application/vnd.devcontainers';
exports.DEVCONTAINER_TAR_LAYER_MEDIATYPE = 'application/vnd.devcontainers.layer.v1+tar';
exports.DEVCONTAINER_COLLECTION_LAYER_MEDIATYPE = 'application/vnd.devcontainers.collection.layer.v1+json';
function getOCIFeatureSet(output, identifier, options, manifest, originalUserFeatureId) {
    const featureRef = getFeatureRef(output, identifier);
    const feat = {
        id: featureRef.id,
        included: true,
        value: options
    };
    const userFeatureIdWithoutVersion = originalUserFeatureId.split(':')[0];
    let featureSet = {
        sourceInformation: {
            type: 'oci',
            manifest: manifest,
            featureRef: featureRef,
            userFeatureId: originalUserFeatureId,
            userFeatureIdWithoutVersion
        },
        features: [feat],
    };
    return featureSet;
}
exports.getOCIFeatureSet = getOCIFeatureSet;
function getFeatureRef(output, resourceAndVersion) {
    // ex: ghcr.io/codspace/features/ruby:1
    const splitOnColon = resourceAndVersion.split(':');
    const resource = splitOnColon[0];
    const version = splitOnColon[1] ? splitOnColon[1] : 'latest';
    const splitOnSlash = resource.split('/');
    const id = splitOnSlash[splitOnSlash.length - 1]; // Aka 'featureName' - Eg: 'ruby'
    const owner = splitOnSlash[1];
    const registry = splitOnSlash[0];
    const namespace = splitOnSlash.slice(1, -1).join('/');
    const path = `${namespace}/${id}`;
    output.write(`resource: ${resource}`, log_1.LogLevel.Trace);
    output.write(`id: ${id}`, log_1.LogLevel.Trace);
    output.write(`version: ${version}`, log_1.LogLevel.Trace);
    output.write(`owner: ${owner}`, log_1.LogLevel.Trace);
    output.write(`namespace: ${namespace}`, log_1.LogLevel.Trace);
    output.write(`registry: ${registry}`, log_1.LogLevel.Trace);
    output.write(`path: ${path}`, log_1.LogLevel.Trace);
    return {
        id,
        version,
        owner,
        namespace,
        registry,
        resource,
        path,
    };
}
exports.getFeatureRef = getFeatureRef;
async function fetchOCIFeatureManifestIfExistsFromUserIdentifier(output, env, identifier, manifestDigest, authToken) {
    const featureRef = getFeatureRef(output, identifier);
    return await fetchOCIFeatureManifestIfExists(output, env, featureRef, manifestDigest, authToken);
}
exports.fetchOCIFeatureManifestIfExistsFromUserIdentifier = fetchOCIFeatureManifestIfExistsFromUserIdentifier;
// Validate if a manifest exists and is reachable about the declared feature.
// Specification: https://github.com/opencontainers/distribution-spec/blob/v1.0.1/spec.md#pulling-manifests
async function fetchOCIFeatureManifestIfExists(output, env, featureRef, manifestDigest, authToken) {
    // Simple mechanism to avoid making a DNS request for 
    // something that is not a domain name.
    if (featureRef.registry.indexOf('.') < 0) {
        return undefined;
    }
    // TODO: Always use the manifest digest (the canonical digest) 
    //       instead of the `featureRef.version` by referencing some lock file (if available).
    let reference = featureRef.version;
    if (manifestDigest) {
        reference = manifestDigest;
    }
    const manifestUrl = `https://${featureRef.registry}/v2/${featureRef.path}/manifests/${reference}`;
    output.write(`manifest url: ${manifestUrl}`, log_1.LogLevel.Trace);
    const manifest = await getFeatureManifest(output, env, manifestUrl, featureRef, authToken);
    if (!manifest) {
        return;
    }
    if ((manifest === null || manifest === void 0 ? void 0 : manifest.config.mediaType) !== exports.DEVCONTAINER_MANIFEST_MEDIATYPE) {
        output.write(`(!) Unexpected manifest media type: ${manifest === null || manifest === void 0 ? void 0 : manifest.config.mediaType}`, log_1.LogLevel.Error);
        return undefined;
    }
    return manifest;
}
exports.fetchOCIFeatureManifestIfExists = fetchOCIFeatureManifestIfExists;
// Download a feature from which a manifest was previously downloaded.
// Specification: https://github.com/opencontainers/distribution-spec/blob/v1.0.1/spec.md#pulling-blobs
async function fetchOCIFeature(output, env, featureSet, ociCacheDir, featCachePath) {
    var _a;
    if (featureSet.sourceInformation.type !== 'oci') {
        output.write(`FeatureSet is not an OCI featureSet.`, log_1.LogLevel.Error);
        throw new Error('FeatureSet is not an OCI featureSet.');
    }
    const { featureRef } = featureSet.sourceInformation;
    const blobUrl = `https://${featureSet.sourceInformation.featureRef.registry}/v2/${featureSet.sourceInformation.featureRef.path}/blobs/${(_a = featureSet.sourceInformation.manifest) === null || _a === void 0 ? void 0 : _a.layers[0].digest}`;
    output.write(`blob url: ${blobUrl}`, log_1.LogLevel.Trace);
    const success = await getFeatureBlob(output, env, blobUrl, ociCacheDir, featCachePath, featureRef);
    if (!success) {
        throw new Error(`Failed to download package for ${featureSet.sourceInformation.featureRef.resource}`);
    }
    return true;
}
exports.fetchOCIFeature = fetchOCIFeature;
async function getFeatureManifest(output, env, url, featureRef, authToken) {
    try {
        const headers = {
            'user-agent': 'devcontainer',
            'accept': 'application/vnd.oci.image.manifest.v1+json',
        };
        const auth = authToken !== null && authToken !== void 0 ? authToken : await fetchRegistryAuthToken(output, featureRef.registry, featureRef.path, env, 'pull');
        if (auth) {
            headers['authorization'] = `Bearer ${auth}`;
        }
        const options = {
            type: 'GET',
            url: url,
            headers: headers
        };
        const response = await (0, httpRequest_1.request)(options);
        const manifest = JSON.parse(response.toString());
        return manifest;
    }
    catch (e) {
        return undefined;
    }
}
exports.getFeatureManifest = getFeatureManifest;
// Downloads a blob from a registry.
async function getFeatureBlob(output, env, url, ociCacheDir, featCachePath, featureRef, authToken) {
    // TODO: Parallelize if multiple layers (not likely).
    // TODO: Seeking might be needed if the size is too large.
    try {
        const tempTarballPath = path.join(ociCacheDir, 'blob.tar');
        const headers = {
            'user-agent': 'devcontainer',
            'accept': 'application/vnd.oci.image.manifest.v1+json',
        };
        const auth = authToken !== null && authToken !== void 0 ? authToken : await fetchRegistryAuthToken(output, featureRef.registry, featureRef.path, env, 'pull');
        if (auth) {
            headers['authorization'] = `Bearer ${auth}`;
        }
        const options = {
            type: 'GET',
            url: url,
            headers: headers
        };
        const blob = await (0, httpRequest_1.request)(options, output);
        await (0, pfs_1.mkdirpLocal)(featCachePath);
        await (0, pfs_1.writeLocalFile)(tempTarballPath, blob);
        await tar.x({
            file: tempTarballPath,
            cwd: featCachePath,
        });
        return true;
    }
    catch (e) {
        output.write(`error: ${e}`, log_1.LogLevel.Error);
        return false;
    }
}
exports.getFeatureBlob = getFeatureBlob;
// https://github.com/oras-project/oras-go/blob/97a9c43c52f9d89ecf5475bc59bd1f96c8cc61f6/registry/remote/auth/scope.go#L60-L74
async function fetchRegistryAuthToken(output, registry, ociRepoPath, env, operationScopes) {
    var _a;
    const headers = {
        'user-agent': 'devcontainer'
    };
    // TODO: Read OS keychain/docker config for auth in various registries!
    let userToken = '';
    if (!!env['GITHUB_TOKEN'] && registry === 'ghcr.io') {
        userToken = env['GITHUB_TOKEN'];
    }
    else if (!!env['DEVCONTAINERS_OCI_AUTH']) {
        // eg: DEVCONTAINERS_OCI_AUTH=domain1:token1,domain2:token2
        const authContexts = env['DEVCONTAINERS_OCI_AUTH'].split(',');
        const authContext = authContexts.find(a => a.split(':')[0] === registry);
        if (authContext && authContext.length === 2) {
            userToken = authContext.split(':')[1];
        }
    }
    else {
        output.write('No oauth authentication credentials found.', log_1.LogLevel.Trace);
    }
    if (userToken) {
        const base64Encoded = Buffer.from(`USERNAME:${userToken}`).toString('base64');
        headers['authorization'] = `Basic ${base64Encoded}`;
    }
    const url = `https://${registry}/token?scope=repo:${ociRepoPath}:${operationScopes}&service=${registry}`;
    output.write(`url: ${url}`, log_1.LogLevel.Trace);
    const options = {
        type: 'GET',
        url: url,
        headers: headers
    };
    let authReq;
    try {
        authReq = await (0, httpRequest_1.request)(options, output);
    }
    catch (e) {
        output.write(`Failed to get registry auth token with error: ${e}`, log_1.LogLevel.Error);
        return undefined;
    }
    if (!authReq) {
        output.write('Failed to get registry auth token', log_1.LogLevel.Error);
        return undefined;
    }
    const token = (_a = JSON.parse(authReq.toString())) === null || _a === void 0 ? void 0 : _a.token;
    if (!token) {
        output.write('Failed to parse registry auth token response', log_1.LogLevel.Error);
        return undefined;
    }
    return token;
}
exports.fetchRegistryAuthToken = fetchRegistryAuthToken;
// Lists published versions/tags of a feature 
// Specification: https://github.com/opencontainers/distribution-spec/blob/v1.0.1/spec.md#content-discovery
async function getPublishedVersions(featureRef, output, sorted = false) {
    var _a;
    try {
        const url = `https://${featureRef.registry}/v2/${featureRef.namespace}/${featureRef.id}/tags/list`;
        let authToken = await fetchRegistryAuthToken(output, featureRef.registry, featureRef.path, process.env, 'pull');
        if (!authToken) {
            output.write(`(!) ERR: Failed to publish feature: ${featureRef.resource}`, log_1.LogLevel.Error);
            return undefined;
        }
        const headers = {
            'user-agent': 'devcontainer',
            'accept': 'application/json',
            'authorization': `Bearer ${authToken}`
        };
        const options = {
            type: 'GET',
            url: url,
            headers: headers
        };
        const response = await (0, httpRequest_1.request)(options);
        const publishedVersionsResponse = JSON.parse(response.toString());
        if (!sorted) {
            return publishedVersionsResponse.tags;
        }
        // Sort tags in descending order, removing latest.
        const hasLatest = publishedVersionsResponse.tags.includes('latest');
        const sortedVersions = publishedVersionsResponse.tags
            .filter(f => f !== 'latest')
            .sort((a, b) => semver.compareIdentifiers(a, b));
        return hasLatest ? ['latest', ...sortedVersions] : sortedVersions;
    }
    catch (e) {
        // Publishing for the first time
        if (e === null || e === void 0 ? void 0 : e.message.includes('HTTP 404: Not Found')) {
            return [];
        }
        output.write(`(!) ERR: Could not fetch published tags for '${featureRef.namespace}/${featureRef.id}' : ${(_a = e === null || e === void 0 ? void 0 : e.message) !== null && _a !== void 0 ? _a : ''} `, log_1.LogLevel.Trace);
        return undefined;
    }
}
exports.getPublishedVersions = getPublishedVersions;
//# sourceMappingURL=containerFeaturesOCI.js.map