"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.calculateManifestAndContentDigest = exports.postUploadSessionId = exports.checkIfBlobExists = exports.calculateDataLayer = exports.generateCompleteManifestForCollectionFile = exports.generateCompleteManifestForIndividualFeature = exports.putBlob = exports.putManifestWithTags = exports.pushFeatureCollectionMetadata = exports.pushOCIFeature = void 0;
const path = __importStar(require("path"));
const fs = __importStar(require("fs"));
const crypto = __importStar(require("crypto"));
const async_1 = require("../spec-common/async");
const httpRequest_1 = require("../spec-utils/httpRequest");
const log_1 = require("../spec-utils/log");
const pfs_1 = require("../spec-utils/pfs");
const containerFeaturesOCI_1 = require("./containerFeaturesOCI");
// (!) Entrypoint function to push a single feature to a registry.
//     Devcontainer Spec : https://containers.dev/implementors/features-distribution/#oci-registry
//     OCI Spec          :  https://github.com/opencontainers/distribution-spec/blob/main/spec.md#push
async function pushOCIFeature(output, featureRef, pathToTgz, tags) {
    output.write(`Starting push of feature '${featureRef.id}' to '${featureRef.resource}' with tags '${tags.join(', ')}'`);
    output.write(`${JSON.stringify(featureRef, null, 2)}`, log_1.LogLevel.Trace);
    const env = process.env;
    // Generate registry auth token with `pull,push` scopes.
    const registryAuthToken = await (0, containerFeaturesOCI_1.fetchRegistryAuthToken)(output, featureRef.registry, featureRef.path, env, 'pull,push');
    if (!registryAuthToken) {
        output.write(`Failed to get registry auth token`, log_1.LogLevel.Error);
        return false;
    }
    // Generate Manifest for given feature artifact.
    const manifest = await generateCompleteManifestForIndividualFeature(output, pathToTgz, featureRef);
    if (!manifest) {
        output.write(`Failed to generate manifest for ${featureRef.id}`, log_1.LogLevel.Error);
        return false;
    }
    output.write(`Generated manifest: \n${JSON.stringify(manifest === null || manifest === void 0 ? void 0 : manifest.manifestObj, undefined, 4)}`, log_1.LogLevel.Trace);
    // If the exact manifest digest already exists in the registry, we don't need to push individual blobs (it's already there!) 
    const existingFeatureManifest = await (0, containerFeaturesOCI_1.fetchOCIFeatureManifestIfExists)(output, env, featureRef, manifest.digest, registryAuthToken);
    if (manifest.digest && existingFeatureManifest) {
        output.write(`Not reuploading blobs, digest already exists.`, log_1.LogLevel.Trace);
        await putManifestWithTags(output, manifest.manifestStr, featureRef, tags, registryAuthToken);
        return true;
    }
    const blobsToPush = [
        {
            name: 'configLayer',
            digest: manifest.manifestObj.config.digest,
        },
        {
            name: 'tgzLayer',
            digest: manifest.manifestObj.layers[0].digest,
        }
    ];
    // Obtain session ID with `/v2/<namespace>/blobs/uploads/` 
    const blobPutLocationUriPath = await postUploadSessionId(output, featureRef, registryAuthToken);
    if (!blobPutLocationUriPath) {
        output.write(`Failed to get upload session ID`, log_1.LogLevel.Error);
        return false;
    }
    for await (const blob of blobsToPush) {
        const { name, digest } = blob;
        const blobExistsConfigLayer = await checkIfBlobExists(output, featureRef, digest, registryAuthToken);
        output.write(`blob: '${name}' with digest '${digest}'  ${blobExistsConfigLayer ? 'already exists' : 'does not exist'} in registry.`, log_1.LogLevel.Trace);
        // PUT blobs
        if (!blobExistsConfigLayer) {
            if (!(await putBlob(output, pathToTgz, blobPutLocationUriPath, featureRef, digest, registryAuthToken))) {
                output.write(`Failed to PUT blob '${name}' with digest '${digest}'`, log_1.LogLevel.Error);
                return false;
            }
        }
    }
    // Send a final PUT to combine blobs and tag manifest properly.
    await putManifestWithTags(output, manifest.manifestStr, featureRef, tags, registryAuthToken);
    // Success!
    return true;
}
exports.pushOCIFeature = pushOCIFeature;
// (!) Entrypoint function to push a collection metadata/overview file for a set of features to a registry.
//     Devcontainer Spec :  https://containers.dev/implementors/features-distribution/#oci-registry (see 'devcontainer-collection.json')
//     OCI Spec          :  https://github.com/opencontainers/distribution-spec/blob/main/spec.md#push
async function pushFeatureCollectionMetadata(output, featureCollectionRef, pathToCollectionJson) {
    output.write(`Starting push of latest feature collection for namespace '${featureCollectionRef.path}' to '${featureCollectionRef.registry}'`);
    output.write(`${JSON.stringify(featureCollectionRef, null, 2)}`, log_1.LogLevel.Trace);
    const env = process.env;
    const registryAuthToken = await (0, containerFeaturesOCI_1.fetchRegistryAuthToken)(output, featureCollectionRef.registry, featureCollectionRef.path, env, 'pull,push');
    if (!registryAuthToken) {
        output.write(`Failed to get registry auth token`, log_1.LogLevel.Error);
        return false;
    }
    // Generate Manifest for collection artifact.
    const manifest = await generateCompleteManifestForCollectionFile(output, pathToCollectionJson, featureCollectionRef);
    if (!manifest) {
        output.write(`Failed to generate manifest for ${featureCollectionRef.path}`, log_1.LogLevel.Error);
        return false;
    }
    output.write(`Generated manifest: \n${JSON.stringify(manifest === null || manifest === void 0 ? void 0 : manifest.manifestObj, undefined, 4)}`, log_1.LogLevel.Trace);
    // If the exact manifest digest already exists in the registry, we don't need to push individual blobs (it's already there!) 
    const existingFeatureManifest = await (0, containerFeaturesOCI_1.fetchOCIFeatureManifestIfExists)(output, env, featureCollectionRef, manifest.digest, registryAuthToken);
    if (manifest.digest && existingFeatureManifest) {
        output.write(`Not reuploading blobs, digest already exists.`, log_1.LogLevel.Trace);
        await putManifestWithTags(output, manifest.manifestStr, featureCollectionRef, ['latest'], registryAuthToken);
        return true;
    }
    // Obtain session ID with `/v2/<namespace>/blobs/uploads/` 
    const blobPutLocationUriPath = await postUploadSessionId(output, featureCollectionRef, registryAuthToken);
    if (!blobPutLocationUriPath) {
        output.write(`Failed to get upload session ID`, log_1.LogLevel.Error);
        return false;
    }
    const blobsToPush = [
        {
            name: 'configLayer',
            digest: manifest.manifestObj.config.digest,
        },
        {
            name: 'collectionLayer',
            digest: manifest.manifestObj.layers[0].digest,
        }
    ];
    for await (const blob of blobsToPush) {
        const { name, digest } = blob;
        const blobExistsConfigLayer = await checkIfBlobExists(output, featureCollectionRef, digest, registryAuthToken);
        output.write(`blob: '${name}' with digest '${digest}'  ${blobExistsConfigLayer ? 'already exists' : 'does not exist'} in registry.`, log_1.LogLevel.Trace);
        // PUT blobs
        if (!blobExistsConfigLayer) {
            if (!(await putBlob(output, pathToCollectionJson, blobPutLocationUriPath, featureCollectionRef, digest, registryAuthToken))) {
                output.write(`Failed to PUT blob '${name}' with digest '${digest}'`, log_1.LogLevel.Error);
                return false;
            }
        }
    }
    // Send a final PUT to combine blobs and tag manifest properly.
    // Collections are always tagged 'latest'
    await putManifestWithTags(output, manifest.manifestStr, featureCollectionRef, ['latest'], registryAuthToken);
    return true;
}
exports.pushFeatureCollectionMetadata = pushFeatureCollectionMetadata;
// --- Helper Functions
// Spec: https://github.com/opencontainers/distribution-spec/blob/main/spec.md#pushing-manifests (PUT /manifests/<ref>)
async function putManifestWithTags(output, manifestStr, featureRef, tags, registryAuthToken) {
    output.write(`Tagging manifest with tags: ${tags.join(', ')}`, log_1.LogLevel.Trace);
    for await (const tag of tags) {
        const url = `https://${featureRef.registry}/v2/${featureRef.path}/manifests/${tag}`;
        output.write(`PUT -> '${url}'`, log_1.LogLevel.Trace);
        const options = {
            type: 'PUT',
            url,
            headers: {
                'Authorization': `Bearer ${registryAuthToken}`,
                'Content-Type': 'application/vnd.oci.image.manifest.v1+json',
            },
            data: Buffer.from(manifestStr),
        };
        let { statusCode, resHeaders } = await (0, httpRequest_1.requestResolveHeaders)(options);
        // Retry logic: when request fails with HTTP 429: too many requests
        if (statusCode === 429) {
            output.write(`Failed to PUT manifest for tag ${tag} due to too many requests. Retrying...`, log_1.LogLevel.Warning);
            await (0, async_1.delay)(2000);
            let response = await (0, httpRequest_1.requestResolveHeaders)(options);
            statusCode = response.statusCode;
            resHeaders = response.resHeaders;
        }
        if (statusCode !== 201) {
            output.write(`Failed to PUT manifest for tag ${tag}`, log_1.LogLevel.Error);
            return false;
        }
        const dockerContentDigestResponseHeader = resHeaders['docker-content-digest'] || resHeaders['Docker-Content-Digest'];
        const locationResponseHeader = resHeaders['location'] || resHeaders['Location'];
        output.write(`Tagged: ${tag} -> ${locationResponseHeader}`, log_1.LogLevel.Info);
        output.write(`Returned Content-Digest: ${dockerContentDigestResponseHeader}`, log_1.LogLevel.Trace);
    }
    return true;
}
exports.putManifestWithTags = putManifestWithTags;
// Spec: https://github.com/opencontainers/distribution-spec/blob/main/spec.md#post-then-put (PUT <location>?digest=<digest>)
async function putBlob(output, pathToBlob, blobPutLocationUriPath, featureRef, digest, registryAuthToken) {
    output.write(`PUT new blob -> '${digest}'`, log_1.LogLevel.Info);
    if (!(await (0, pfs_1.isLocalFile)(pathToBlob))) {
        output.write(`Blob ${pathToBlob} does not exist`, log_1.LogLevel.Error);
        return false;
    }
    const headers = {
        'user-agent': 'devcontainer',
        'authorization': `Bearer ${registryAuthToken}`,
        'content-type': 'application/octet-stream',
    };
    // OCI distribution spec is ambiguous on whether we get back an absolute or relative path.
    let url = '';
    if (blobPutLocationUriPath.startsWith('https://')) {
        url = blobPutLocationUriPath;
    }
    else {
        url = `https://${featureRef.registry}${blobPutLocationUriPath}`;
    }
    url += `?digest=${digest}`;
    output.write(`Crafted blob url:  ${url}`, log_1.LogLevel.Trace);
    const { statusCode } = await (0, httpRequest_1.requestResolveHeaders)({ type: 'PUT', url, headers, data: await (0, pfs_1.readLocalFile)(pathToBlob) });
    if (statusCode !== 201) {
        output.write(`${statusCode}: Failed to upload blob '${pathToBlob}' to '${url}'`, log_1.LogLevel.Error);
        return false;
    }
    return true;
}
exports.putBlob = putBlob;
// Generate a layer that follows the `application/vnd.devcontainers.layer.v1+tar` mediaType
// as defined in: https://containers.dev/implementors/features-distribution/#oci-registry
async function generateCompleteManifestForIndividualFeature(output, pathToTgz, ociFeatureRef) {
    const tgzLayer = await calculateDataLayer(output, pathToTgz, containerFeaturesOCI_1.DEVCONTAINER_TAR_LAYER_MEDIATYPE);
    if (!tgzLayer) {
        output.write(`Failed to calculate tgz layer.`, log_1.LogLevel.Error);
        return undefined;
    }
    let annotations = undefined;
    // Specific registries look for certain optional metadata 
    // in the manifest, in this case for UI presentation.
    if (ociFeatureRef.registry === 'ghcr.io') {
        annotations = {
            'com.github.package.type': 'devcontainer_feature',
        };
    }
    return await calculateManifestAndContentDigest(output, tgzLayer, annotations);
}
exports.generateCompleteManifestForIndividualFeature = generateCompleteManifestForIndividualFeature;
// Generate a layer that follows the `application/vnd.devcontainers.collection.layer.v1+json` mediaType
// as defined in: https://containers.dev/implementors/features-distribution/#oci-registry
async function generateCompleteManifestForCollectionFile(output, pathToCollectionFile, collectionRef) {
    const collectionMetadataLayer = await calculateDataLayer(output, pathToCollectionFile, containerFeaturesOCI_1.DEVCONTAINER_COLLECTION_LAYER_MEDIATYPE);
    if (!collectionMetadataLayer) {
        output.write(`Failed to calculate collection file layer.`, log_1.LogLevel.Error);
        return undefined;
    }
    let annotations = undefined;
    // Specific registries look for certain optional metadata 
    // in the manifest, in this case for UI presentation.
    if (collectionRef.registry === 'ghcr.io') {
        annotations = {
            'com.github.package.type': 'devcontainer_collection',
        };
    }
    return await calculateManifestAndContentDigest(output, collectionMetadataLayer, annotations);
}
exports.generateCompleteManifestForCollectionFile = generateCompleteManifestForCollectionFile;
// Generic construction of a layer in the manifest and digest for the generated layer.
async function calculateDataLayer(output, pathToData, mediaType) {
    output.write(`Creating manifest from ${pathToData}`, log_1.LogLevel.Trace);
    if (!(await (0, pfs_1.isLocalFile)(pathToData))) {
        output.write(`${pathToData} does not exist.`, log_1.LogLevel.Error);
        return undefined;
    }
    const dataBytes = fs.readFileSync(pathToData);
    const tarSha256 = crypto.createHash('sha256').update(dataBytes).digest('hex');
    output.write(`${pathToData}:  sha256:${tarSha256} (size: ${dataBytes.byteLength})`, log_1.LogLevel.Info);
    return {
        mediaType,
        digest: `sha256:${tarSha256}`,
        size: dataBytes.byteLength,
        annotations: {
            'org.opencontainers.image.title': path.basename(pathToData),
        }
    };
}
exports.calculateDataLayer = calculateDataLayer;
// Spec: https://github.com/opencontainers/distribution-spec/blob/main/spec.md#checking-if-content-exists-in-the-registry
//       Requires registry auth token.
async function checkIfBlobExists(output, featureRef, digest, authToken) {
    const headers = {
        'user-agent': 'devcontainer',
        'authorization': `Bearer ${authToken}`,
    };
    const url = `https://${featureRef.registry}/v2/${featureRef.path}/blobs/${digest}`;
    const statusCode = await (0, httpRequest_1.headRequest)({ url, headers }, output);
    output.write(`${url}: ${statusCode}`, log_1.LogLevel.Trace);
    return statusCode === 200;
}
exports.checkIfBlobExists = checkIfBlobExists;
// Spec: https://github.com/opencontainers/distribution-spec/blob/main/spec.md#post-then-put
//       Requires registry auth token.
async function postUploadSessionId(output, featureRef, authToken) {
    const headers = {
        'user-agent': 'devcontainer',
        'authorization': `Bearer ${authToken}`,
    };
    const url = `https://${featureRef.registry}/v2/${featureRef.path}/blobs/uploads/`;
    output.write(`Generating Upload URL -> ${url}`, log_1.LogLevel.Trace);
    const { statusCode, resHeaders } = await (0, httpRequest_1.requestResolveHeaders)({ type: 'POST', url, headers }, output);
    output.write(`${url}: ${statusCode}`, log_1.LogLevel.Trace);
    if (statusCode === 202) {
        const locationHeader = resHeaders['location'] || resHeaders['Location'];
        if (!locationHeader) {
            output.write(`${url}: Got 202 status code, but no location header found.`, log_1.LogLevel.Error);
            return undefined;
        }
        return locationHeader;
    }
    return undefined;
}
exports.postUploadSessionId = postUploadSessionId;
async function calculateManifestAndContentDigest(output, dataLayer, annotations) {
    // A canonical manifest digest is the sha256 hash of the JSON representation of the manifest, without the signature content.
    // See: https://docs.docker.com/registry/spec/api/#content-digests
    // Below is an example of a serialized manifest that should resolve to '9726054859c13377c4c3c3c73d15065de59d0c25d61d5652576c0125f2ea8ed3'
    // {"schemaVersion":2,"mediaType":"application/vnd.oci.image.manifest.v1+json","config":{"mediaType":"application/vnd.devcontainers","digest":"sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855","size":0},"layers":[{"mediaType":"application/vnd.devcontainers.layer.v1+tar","digest":"sha256:b2006e7647191f7b47222ae48df049c6e21a4c5a04acfad0c4ef614d819de4c5","size":15872,"annotations":{"org.opencontainers.image.title":"go.tgz"}}]}
    let manifest = {
        schemaVersion: 2,
        mediaType: 'application/vnd.oci.image.manifest.v1+json',
        config: {
            mediaType: 'application/vnd.devcontainers',
            digest: 'sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',
            size: 0
        },
        layers: [
            dataLayer
        ],
    };
    if (annotations) {
        manifest.annotations = annotations;
    }
    const manifestStringified = JSON.stringify(manifest);
    const manifestHash = crypto.createHash('sha256').update(manifestStringified).digest('hex');
    output.write(`Computed Content-Digest ->  sha256:${manifestHash} (size: ${manifestHash.length})`, log_1.LogLevel.Info);
    return {
        manifestStr: manifestStringified,
        manifestObj: manifest,
        digest: manifestHash,
    };
}
exports.calculateManifestAndContentDigest = calculateManifestAndContentDigest;
//# sourceMappingURL=containerFeaturesOCIPush.js.map