"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PackageOptions = void 0;
const targetPositionalDescription = (collectionType) => `
Package ${collectionType}s at provided [target] (default is cwd), where [target] is either:
   1. A path to the src folder of the collection with [1..n] ${collectionType}s.
   2. A path to a single ${collectionType} that contains a devcontainer-${collectionType}.json.
   
   Additionally, a 'devcontainer-collection.json' will be generated in the output directory.
`;
function PackageOptions(y, collectionType) {
    return y
        .options({
        'output-folder': { type: 'string', alias: 'o', default: './output', description: 'Path to output directory. Will create directories as needed.' },
        'force-clean-output-folder': { type: 'boolean', alias: 'f', default: false, description: 'Automatically delete previous output directory before packaging' },
        'log-level': { choices: ['info', 'debug', 'trace'], default: 'info', description: 'Log level.' },
    })
        .positional('target', { type: 'string', default: '.', description: targetPositionalDescription(collectionType) })
        .check(_argv => {
        return true;
    });
}
exports.PackageOptions = PackageOptions;
//# sourceMappingURL=package.js.map