"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateRemoteUserUID = exports.getRemoteUserUIDUpdateDetails = exports.generateContainerEnvs = exports.getExtendImageBuildInfo = exports.extendImage = exports.getSafeId = void 0;
const path = __importStar(require("path"));
const string_decoder_1 = require("string_decoder");
const tar = __importStar(require("tar"));
const dockerUtils_1 = require("../spec-shutdown/dockerUtils");
const log_1 = require("../spec-utils/log");
const containerFeaturesConfiguration_1 = require("../spec-configuration/containerFeaturesConfiguration");
const pfs_1 = require("../spec-utils/pfs");
const product_1 = require("../spec-utils/product");
const utils_1 = require("./utils");
const commonUtils_1 = require("../spec-common/commonUtils");
const imageMetadata_1 = require("./imageMetadata");
// Escapes environment variable keys.
//
// Environment variables must contain:
//      - alpha-numeric values, or
//      - the '_' character, and
//      - a number cannot be the first character 
const getSafeId = (str) => str
    .replace(/[^\w_]/g, '_')
    .replace(/^[\d_]+/g, '_')
    .toUpperCase();
exports.getSafeId = getSafeId;
async function extendImage(params, config, imageName) {
    const { common } = params;
    const { cliHost, output } = common;
    const imageBuildInfo = await (0, imageMetadata_1.getImageBuildInfoFromImage)(params, imageName, config.substitute, common.experimentalImageMetadata);
    const extendImageDetails = await getExtendImageBuildInfo(params, config, imageName, imageBuildInfo);
    if (!extendImageDetails || !extendImageDetails.featureBuildInfo) {
        // no feature extensions - return
        return {
            updatedImageName: [imageName],
            imageMetadata: imageBuildInfo.metadata,
            imageDetails: async () => imageBuildInfo.imageDetails,
        };
    }
    const { featureBuildInfo, collapsedFeaturesConfig } = extendImageDetails;
    // Got feature extensions -> build the image
    const dockerfilePath = cliHost.path.join(featureBuildInfo.dstFolder, 'Dockerfile.extended');
    await cliHost.writeFile(dockerfilePath, Buffer.from(featureBuildInfo.dockerfilePrefixContent + featureBuildInfo.dockerfileContent));
    const folderImageName = (0, utils_1.getFolderImageName)(common);
    const updatedImageName = `${imageName.startsWith(folderImageName) ? imageName : folderImageName}-features`;
    const args = [];
    if (params.buildKitVersion) {
        args.push('buildx', 'build', '--load');
        for (const buildContext in featureBuildInfo.buildKitContexts) {
            args.push('--build-context', `${buildContext}=${featureBuildInfo.buildKitContexts[buildContext]}`);
        }
    }
    else {
        args.push('build');
    }
    for (const buildArg in featureBuildInfo.buildArgs) {
        args.push('--build-arg', `${buildArg}=${featureBuildInfo.buildArgs[buildArg]}`);
    }
    // Once this is step merged with the user Dockerfile (or working against the base image),
    // the path will be the dev container context
    // Set empty dir under temp path as the context for now to ensure we don't have dependencies on the features content
    const emptyTempDir = (0, utils_1.getEmptyContextFolder)(common);
    cliHost.mkdirp(emptyTempDir);
    args.push('--target', featureBuildInfo.overrideTarget, '-t', updatedImageName, '-f', dockerfilePath, emptyTempDir);
    if (params.isTTY) {
        const infoParams = { ...(0, dockerUtils_1.toPtyExecParameters)(params), output: (0, log_1.makeLog)(output, log_1.LogLevel.Info) };
        await (0, dockerUtils_1.dockerPtyCLI)(infoParams, ...args);
    }
    else {
        const infoParams = { ...(0, dockerUtils_1.toExecParameters)(params), output: (0, log_1.makeLog)(output, log_1.LogLevel.Info), print: 'continuous' };
        await (0, dockerUtils_1.dockerCLI)(infoParams, ...args);
    }
    return {
        updatedImageName: [updatedImageName],
        imageMetadata: (0, imageMetadata_1.getDevcontainerMetadata)(imageBuildInfo.metadata, config, (collapsedFeaturesConfig === null || collapsedFeaturesConfig === void 0 ? void 0 : collapsedFeaturesConfig.allFeatures) || []),
        imageDetails: async () => imageBuildInfo.imageDetails,
    };
}
exports.extendImage = extendImage;
async function getExtendImageBuildInfo(params, config, baseName, imageBuildInfo) {
    // Creates the folder where the working files will be setup.
    const dstFolder = await (0, utils_1.createFeaturesTempFolder)(params.common);
    // Extracts the local cache of features.
    await createLocalFeatures(params, dstFolder);
    // Processes the user's configuration.
    const featuresConfig = await (0, containerFeaturesConfiguration_1.generateFeaturesConfig)(params.common, dstFolder, config.config, containerFeaturesConfiguration_1.getContainerFeaturesFolder);
    if (!featuresConfig) {
        return { featureBuildInfo: getImageBuildOptions(params, config, dstFolder, baseName, imageBuildInfo) };
    }
    // Generates the end configuration.
    const collapsedFeaturesConfig = (0, containerFeaturesConfiguration_1.collapseFeaturesConfig)(featuresConfig);
    const featureBuildInfo = await getFeaturesBuildOptions(params, config, featuresConfig, baseName, imageBuildInfo);
    if (!featureBuildInfo) {
        return undefined;
    }
    return { featureBuildInfo, collapsedFeaturesConfig };
}
exports.getExtendImageBuildInfo = getExtendImageBuildInfo;
// NOTE: only exported to enable testing. Not meant to be called outside file.
function generateContainerEnvs(featuresConfig) {
    let result = '';
    for (const fSet of featuresConfig.featureSets) {
        // We only need to generate this ENV references for the initial features specification.
        if (fSet.internalVersion !== '2') {
            result += '\n';
            result += fSet.features
                .filter(f => (product_1.includeAllConfiguredFeatures || f.included) && f.value)
                .reduce((envs, f) => envs.concat(Object.keys(f.containerEnv || {})
                .map(k => `ENV ${k}=${f.containerEnv[k]}`)), [])
                .join('\n');
        }
    }
    return result;
}
exports.generateContainerEnvs = generateContainerEnvs;
async function createLocalFeatures(params, dstFolder) {
    const { common } = params;
    const { cliHost, output } = common;
    // Name of the local cache folder inside the working directory
    const localCacheBuildFolderName = 'local-cache';
    const srcFolder = (0, containerFeaturesConfiguration_1.getContainerFeaturesFolder)(common.extensionPath);
    output.write(`local container features stored at: ${srcFolder}`);
    await cliHost.mkdirp(`${dstFolder}/${localCacheBuildFolderName}`);
    const create = tar.c({
        cwd: srcFolder,
        filter: path => (path !== './Dockerfile' && path !== `./${containerFeaturesConfiguration_1.V1_DEVCONTAINER_FEATURES_FILE_NAME}`),
    }, ['.']);
    const createExit = new Promise((resolve, reject) => {
        create.on('error', reject);
        create.on('finish', resolve);
    });
    const extract = await cliHost.exec({
        cmd: 'tar',
        args: [
            '--no-same-owner',
            '-x',
            '-f', '-',
        ],
        cwd: `${dstFolder}/${localCacheBuildFolderName}`,
        output,
    });
    const stdoutDecoder = new string_decoder_1.StringDecoder();
    extract.stdout.on('data', (chunk) => {
        output.write(stdoutDecoder.write(chunk));
    });
    const stderrDecoder = new string_decoder_1.StringDecoder();
    extract.stderr.on('data', (chunk) => {
        output.write((0, log_1.toErrorText)(stderrDecoder.write(chunk)));
    });
    create.pipe(extract.stdin);
    await extract.exit;
    await createExit; // Allow errors to surface.
}
function getImageBuildOptions(params, config, dstFolder, baseName, imageBuildInfo) {
    return {
        dstFolder,
        dockerfileContent: `
FROM $_DEV_CONTAINERS_BASE_IMAGE AS dev_containers_target_stage
${(0, imageMetadata_1.getDevcontainerMetadataLabel)(imageBuildInfo.metadata, config, [], params.common.experimentalImageMetadata)}
`,
        overrideTarget: 'dev_containers_target_stage',
        dockerfilePrefixContent: `
ARG _DEV_CONTAINERS_BASE_IMAGE=placeholder
`,
        buildArgs: {
            _DEV_CONTAINERS_BASE_IMAGE: baseName,
        },
        buildKitContexts: {},
    };
}
async function getFeaturesBuildOptions(params, devContainerConfig, featuresConfig, baseName, imageBuildInfo) {
    const { common } = params;
    const { cliHost, output } = common;
    const { dstFolder } = featuresConfig;
    if (!dstFolder || dstFolder === '') {
        output.write('dstFolder is undefined or empty in addContainerFeatures', log_1.LogLevel.Error);
        return undefined;
    }
    const buildStageScripts = await Promise.all(featuresConfig.featureSets
        .map(featureSet => containerFeaturesConfiguration_1.multiStageBuildExploration ? featureSet.features
        .filter(f => (product_1.includeAllConfiguredFeatures || f.included) && f.value)
        .reduce(async (binScripts, feature) => {
        const binPath = cliHost.path.join(dstFolder, (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'features', feature.id, 'bin');
        const hasAcquire = cliHost.isFile(cliHost.path.join(binPath, 'acquire'));
        const hasConfigure = cliHost.isFile(cliHost.path.join(binPath, 'configure'));
        const map = await binScripts;
        map[feature.id] = {
            hasAcquire: await hasAcquire,
            hasConfigure: await hasConfigure,
        };
        return map;
    }, Promise.resolve({})) : Promise.resolve({})));
    // With Buildkit (0.8.0 or later), we can supply an additional build context to provide access to
    // the container-features content.
    // For non-Buildkit, we build a temporary image to hold the container-features content in a way
    // that is accessible from the docker build for non-BuiltKit builds
    // TODO generate an image name that is specific to this dev container?
    const buildKitVersionParsed = params.buildKitVersion ? (0, commonUtils_1.parseVersion)(params.buildKitVersion) : null;
    const minRequiredVersion = [0, 8, 0];
    const useBuildKitBuildContexts = buildKitVersionParsed ? !(0, commonUtils_1.isEarlierVersion)(buildKitVersionParsed, minRequiredVersion) : false;
    const buildContentImageName = 'dev_container_feature_content_temp';
    // When copying via buildkit, the content is accessed via '.' (i.e. in the context root)
    // When copying via temp image, the content is in '/tmp/build-features'
    const contentSourceRootPath = useBuildKitBuildContexts ? '.' : '/tmp/build-features/';
    const dockerfile = (0, containerFeaturesConfiguration_1.getContainerFeaturesBaseDockerFile)()
        .replace('#{nonBuildKitFeatureContentFallback}', useBuildKitBuildContexts ? '' : `FROM ${buildContentImageName} as dev_containers_feature_content_source`)
        .replace('{contentSourceRootPath}', contentSourceRootPath)
        .replace('#{featureBuildStages}', getFeatureBuildStages(featuresConfig, buildStageScripts, contentSourceRootPath))
        .replace('#{featureLayer}', (0, containerFeaturesConfiguration_1.getFeatureLayers)(featuresConfig))
        .replace('#{containerEnv}', generateContainerEnvs(featuresConfig))
        .replace('#{copyFeatureBuildStages}', getCopyFeatureBuildStages(featuresConfig, buildStageScripts))
        .replace('#{devcontainerMetadata}', (0, imageMetadata_1.getDevcontainerMetadataLabel)(imageBuildInfo.metadata, devContainerConfig, featuresConfig, common.experimentalImageMetadata));
    const dockerfilePrefixContent = `${useBuildKitBuildContexts ? '# syntax=docker/dockerfile:1.4' : ''}
ARG _DEV_CONTAINERS_BASE_IMAGE=placeholder
`;
    // Build devcontainer-features.env and devcontainer-features-install.sh file(s) for each features source folder
    for await (const fSet of featuresConfig.featureSets) {
        let i = 0;
        if (fSet.internalVersion === '2') {
            for await (const fe of fSet.features) {
                if (fe.cachePath) {
                    fe.internalVersion = '2';
                    const envPath = cliHost.path.join(fe.cachePath, 'devcontainer-features.env');
                    const variables = getFeatureEnvVariables(fe);
                    await cliHost.writeFile(envPath, Buffer.from(variables.join('\n')));
                    const installWrapperPath = cliHost.path.join(fe.cachePath, 'devcontainer-features-install.sh');
                    const installWrapperContent = (0, containerFeaturesConfiguration_1.getFeatureInstallWrapperScript)(fe, fSet, variables);
                    await cliHost.writeFile(installWrapperPath, Buffer.from(installWrapperContent));
                }
            }
        }
        else {
            const featuresEnv = [].concat(...fSet.features
                .filter(f => { var _a; return (product_1.includeAllConfiguredFeatures || f.included) && f.value && !((_a = buildStageScripts[i][f.id]) === null || _a === void 0 ? void 0 : _a.hasAcquire); })
                .map(getFeatureEnvVariables)).join('\n');
            const envPath = cliHost.path.join(fSet.features[0].cachePath, 'devcontainer-features.env');
            await Promise.all([
                cliHost.writeFile(envPath, Buffer.from(featuresEnv)),
                ...fSet.features
                    .filter(f => { var _a; return (product_1.includeAllConfiguredFeatures || f.included) && f.value && ((_a = buildStageScripts[i][f.id]) === null || _a === void 0 ? void 0 : _a.hasAcquire); })
                    .map(f => {
                    const featuresEnv = [
                        ...getFeatureEnvVariables(f),
                        `_BUILD_ARG_${(0, exports.getSafeId)(f.id)}_TARGETPATH=${path.posix.join('/usr/local/devcontainer-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(fSet.sourceInformation), f.id)}`
                    ]
                        .join('\n');
                    const envPath = cliHost.path.join(dstFolder, (0, containerFeaturesConfiguration_1.getSourceInfoString)(fSet.sourceInformation), 'features', f.id, 'devcontainer-features.env'); // next to bin/acquire
                    return cliHost.writeFile(envPath, Buffer.from(featuresEnv));
                })
            ]);
        }
        i++;
    }
    // For non-BuildKit, build the temporary image for the container-features content
    if (!useBuildKitBuildContexts) {
        const buildContentDockerfile = `
	FROM scratch
	COPY . /tmp/build-features/
	`;
        const buildContentDockerfilePath = cliHost.path.join(dstFolder, 'Dockerfile.buildContent');
        await cliHost.writeFile(buildContentDockerfilePath, Buffer.from(buildContentDockerfile));
        const buildContentArgs = [
            'build',
            '-t', buildContentImageName,
            '-f', buildContentDockerfilePath,
        ];
        buildContentArgs.push(dstFolder);
        if (params.isTTY) {
            const buildContentInfoParams = { ...(0, dockerUtils_1.toPtyExecParameters)(params), output: (0, log_1.makeLog)(output, log_1.LogLevel.Info) };
            await (0, dockerUtils_1.dockerPtyCLI)(buildContentInfoParams, ...buildContentArgs);
        }
        else {
            const buildContentInfoParams = { ...(0, dockerUtils_1.toExecParameters)(params), output: (0, log_1.makeLog)(output, log_1.LogLevel.Info), print: 'continuous' };
            await (0, dockerUtils_1.dockerCLI)(buildContentInfoParams, ...buildContentArgs);
        }
    }
    return {
        dstFolder,
        dockerfileContent: dockerfile,
        overrideTarget: 'dev_containers_target_stage',
        dockerfilePrefixContent,
        buildArgs: {
            _DEV_CONTAINERS_BASE_IMAGE: baseName,
            _DEV_CONTAINERS_IMAGE_USER: imageBuildInfo.user,
            _DEV_CONTAINERS_FEATURE_CONTENT_SOURCE: buildContentImageName,
        },
        buildKitContexts: useBuildKitBuildContexts ? { dev_containers_feature_content_source: dstFolder } : {},
    };
}
function getFeatureBuildStages(featuresConfig, buildStageScripts, contentSourceRootPath) {
    return [].concat(...featuresConfig.featureSets
        .map((featureSet, i) => featureSet.features
        .filter(f => { var _a; return (product_1.includeAllConfiguredFeatures || f.included) && f.value && ((_a = buildStageScripts[i][f.id]) === null || _a === void 0 ? void 0 : _a.hasAcquire); })
        .map(f => `FROM mcr.microsoft.com/vscode/devcontainers/base:0-focal as ${(0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation)}_${f.id}
COPY --from=dev_containers_feature_content_source ${path.posix.join(contentSourceRootPath, (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'features', f.id)} ${path.posix.join('/tmp/build-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'features', f.id)}
COPY --from=dev_containers_feature_content_source ${path.posix.join(contentSourceRootPath, (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'common')} ${path.posix.join('/tmp/build-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'common')}
RUN cd ${path.posix.join('/tmp/build-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'features', f.id)} && set -a && . ./devcontainer-features.env && set +a && ./bin/acquire`))).join('\n\n');
}
function getCopyFeatureBuildStages(featuresConfig, buildStageScripts) {
    return [].concat(...featuresConfig.featureSets
        .map((featureSet, i) => featureSet.features
        .filter(f => { var _a; return (product_1.includeAllConfiguredFeatures || f.included) && f.value && ((_a = buildStageScripts[i][f.id]) === null || _a === void 0 ? void 0 : _a.hasAcquire); })
        .map(f => {
        var _a;
        const featurePath = path.posix.join('/usr/local/devcontainer-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), f.id);
        return `COPY --from=${(0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation)}_${f.id} ${featurePath} ${featurePath}${((_a = buildStageScripts[i][f.id]) === null || _a === void 0 ? void 0 : _a.hasConfigure) ? `
RUN cd ${path.posix.join('/tmp/build-features', (0, containerFeaturesConfiguration_1.getSourceInfoString)(featureSet.sourceInformation), 'features', f.id)} && set -a && . ./devcontainer-features.env && set +a && ./bin/configure` : ''}`;
    }))).join('\n\n');
}
function getFeatureEnvVariables(f) {
    const values = (0, containerFeaturesConfiguration_1.getFeatureValueObject)(f);
    const idSafe = (0, exports.getSafeId)(f.id);
    const variables = [];
    if (f.internalVersion !== '2') {
        if (values) {
            variables.push(...Object.keys(values)
                .map(name => `_BUILD_ARG_${idSafe}_${(0, exports.getSafeId)(name)}="${values[name]}"`));
            variables.push(`_BUILD_ARG_${idSafe}=true`);
        }
        if (f.buildArg) {
            variables.push(`${f.buildArg}=${(0, containerFeaturesConfiguration_1.getFeatureMainValue)(f)}`);
        }
        return variables;
    }
    else {
        if (values) {
            variables.push(...Object.keys(values)
                .map(name => `${(0, exports.getSafeId)(name)}="${values[name]}"`));
        }
        if (f.buildArg) {
            variables.push(`${f.buildArg}=${(0, containerFeaturesConfiguration_1.getFeatureMainValue)(f)}`);
        }
        return variables;
    }
}
async function getRemoteUserUIDUpdateDetails(params, mergedConfig, imageName, imageDetails, runArgsUser) {
    const { common } = params;
    const { cliHost } = common;
    const { updateRemoteUserUID } = mergedConfig;
    if (params.updateRemoteUserUIDDefault === 'never' || !(typeof updateRemoteUserUID === 'boolean' ? updateRemoteUserUID : params.updateRemoteUserUIDDefault === 'on') || !(cliHost.platform === 'linux' || params.updateRemoteUserUIDOnMacOS && cliHost.platform === 'darwin')) {
        return null;
    }
    const details = await imageDetails();
    const imageUser = details.Config.User || 'root';
    const remoteUser = mergedConfig.remoteUser || runArgsUser || imageUser;
    if (remoteUser === 'root' || /^\d+$/.test(remoteUser)) {
        return null;
    }
    const folderImageName = (0, utils_1.getFolderImageName)(common);
    const fixedImageName = `${imageName.startsWith(folderImageName) ? imageName : folderImageName}-uid`;
    return {
        imageName: fixedImageName,
        remoteUser,
        imageUser,
    };
}
exports.getRemoteUserUIDUpdateDetails = getRemoteUserUIDUpdateDetails;
async function updateRemoteUserUID(params, mergedConfig, imageName, imageDetails, runArgsUser) {
    const { common } = params;
    const { cliHost } = common;
    const updateDetails = await getRemoteUserUIDUpdateDetails(params, mergedConfig, imageName, imageDetails, runArgsUser);
    if (!updateDetails) {
        return imageName;
    }
    const { imageName: fixedImageName, remoteUser, imageUser } = updateDetails;
    const dockerfileName = 'updateUID.Dockerfile';
    const srcDockerfile = path.join(common.extensionPath, 'scripts', dockerfileName);
    const version = common.package.version;
    const destDockerfile = cliHost.path.join(await (0, utils_1.getCacheFolder)(cliHost), `${dockerfileName}-${version}`);
    const tmpDockerfile = `${destDockerfile}-${Date.now()}`;
    await cliHost.mkdirp(cliHost.path.dirname(tmpDockerfile));
    await cliHost.writeFile(tmpDockerfile, await (0, pfs_1.readLocalFile)(srcDockerfile));
    await cliHost.rename(tmpDockerfile, destDockerfile);
    const emptyFolder = (0, utils_1.getEmptyContextFolder)(common);
    await cliHost.mkdirp(emptyFolder);
    const args = [
        'build',
        '-f', destDockerfile,
        '-t', fixedImageName,
        '--build-arg', `BASE_IMAGE=${imageName}`,
        '--build-arg', `REMOTE_USER=${remoteUser}`,
        '--build-arg', `NEW_UID=${await cliHost.getuid()}`,
        '--build-arg', `NEW_GID=${await cliHost.getgid()}`,
        '--build-arg', `IMAGE_USER=${imageUser}`,
        emptyFolder,
    ];
    if (params.isTTY) {
        await (0, dockerUtils_1.dockerPtyCLI)(params, ...args);
    }
    else {
        await (0, dockerUtils_1.dockerCLI)(params, ...args);
    }
    return fixedImageName;
}
exports.updateRemoteUserUID = updateRemoteUserUID;
//# sourceMappingURL=containerFeatures.js.map