"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.featuresInfoHandler = exports.featuresInfoOptions = void 0;
const containerFeaturesOCI_1 = require("../../spec-configuration/containerFeaturesOCI");
const log_1 = require("../../spec-utils/log");
const product_1 = require("../../spec-utils/product");
const devContainers_1 = require("../devContainers");
function featuresInfoOptions(y) {
    return y
        .options({
        'log-level': { choices: ['info', 'debug', 'trace'], default: 'info', description: 'Log level.' },
        'output-format': { choices: ['text', 'json'], default: 'text', description: 'Output format.' },
    })
        .positional('featureId', { type: 'string', demandOption: true, description: 'Feature Id' });
}
exports.featuresInfoOptions = featuresInfoOptions;
function featuresInfoHandler(args) {
    (async () => await featuresInfo(args))().catch(console.error);
}
exports.featuresInfoHandler = featuresInfoHandler;
async function featuresInfo({ 'featureId': featureId, 'log-level': inputLogLevel, 'output-format': outputFormat, }) {
    const disposables = [];
    const dispose = async () => {
        await Promise.all(disposables.map(d => d()));
    };
    const pkg = (0, product_1.getPackageConfig)();
    const output = (0, devContainers_1.createLog)({
        logLevel: (0, log_1.mapLogLevel)(inputLogLevel),
        logFormat: 'text',
        log: (str) => process.stdout.write(str),
        terminalDimensions: undefined,
    }, pkg, new Date(), disposables, true);
    const featureOciRef = (0, containerFeaturesOCI_1.getFeatureRef)(output, featureId);
    const publishedVersions = await (0, containerFeaturesOCI_1.getPublishedVersions)(featureOciRef, output, true);
    if (!publishedVersions || publishedVersions.length === 0) {
        if (outputFormat === 'json') {
            output.raw(JSON.stringify({}), log_1.LogLevel.Info);
        }
        else {
            output.raw(`No published versions found for feature '${featureId}'\n`, log_1.LogLevel.Error);
        }
        process.exit(1);
    }
    const data = {
        publishedVersions
    };
    if (outputFormat === 'json') {
        printAsJson(output, data);
    }
    else {
        printAsPlainText(output, data);
    }
    await dispose();
    process.exit(0);
}
function printAsJson(output, data) {
    output.raw(JSON.stringify(data, null, 2), log_1.LogLevel.Info);
}
function printAsPlainText(output, data) {
    output.raw(`Published Versions: \n   ${data.publishedVersions.join('\n   ')}\n`, log_1.LogLevel.Info);
}
//# sourceMappingURL=info.js.map