"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.packageFeatureCollection = exports.packageSingleFeature = exports.getFeatureArchiveName = exports.doFeaturesPackageCommand = void 0;
const path_1 = __importDefault(require("path"));
const log_1 = require("../../spec-utils/log");
const pfs_1 = require("../../spec-utils/pfs");
const packageCommandImpl_1 = require("../collectionCommonUtils/packageCommandImpl");
const collectionType = 'feature';
async function doFeaturesPackageCommand(args) {
    args = await (0, packageCommandImpl_1.prepPackageCommand)(args, collectionType);
    const { output, outputDir } = args;
    const isSingleFeature = args.isSingle;
    // For each feature, package each feature and write to 'outputDir/{f}.tgz'
    // Returns an array of feature metadata from each processed feature
    let metadataOutput = [];
    if (isSingleFeature) {
        // Package individual features
        output.write('Packaging single feature...', log_1.LogLevel.Info);
        metadataOutput = await packageSingleFeature(args);
    }
    else {
        output.write('Packaging feature collection...', log_1.LogLevel.Info);
        metadataOutput = await packageFeatureCollection(args);
    }
    if (!metadataOutput) {
        output.write('Failed to package features', log_1.LogLevel.Error);
        return undefined;
    }
    const collection = {
        sourceInformation: {
            source: 'devcontainer-cli',
        },
        features: metadataOutput,
    };
    // Write the metadata to a file
    const metadataOutputPath = path_1.default.join(outputDir, packageCommandImpl_1.OCICollectionFileName);
    await (0, pfs_1.writeLocalFile)(metadataOutputPath, JSON.stringify(collection, null, 4));
    return collection;
}
exports.doFeaturesPackageCommand = doFeaturesPackageCommand;
const getFeatureArchiveName = (f) => `devcontainer-feature-${f}.tgz`;
exports.getFeatureArchiveName = getFeatureArchiveName;
async function packageSingleFeature(args) {
    return await (0, packageCommandImpl_1.packageSingleFeatureOrTemplate)(args, collectionType);
}
exports.packageSingleFeature = packageSingleFeature;
async function packageFeatureCollection(args) {
    return await (0, packageCommandImpl_1.packageCollection)(args, collectionType);
}
exports.packageFeatureCollection = packageFeatureCollection;
//# sourceMappingURL=packageCommandImpl.js.map