"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.featuresPublishHandler = exports.featuresPublishOptions = void 0;
const path_1 = __importDefault(require("path"));
const os = __importStar(require("os"));
const log_1 = require("../../spec-utils/log");
const pfs_1 = require("../../spec-utils/pfs");
const product_1 = require("../../spec-utils/product");
const devContainers_1 = require("../devContainers");
const packageCommandImpl_1 = require("./packageCommandImpl");
const publishCommandImpl_1 = require("./publishCommandImpl");
const containerFeaturesOCI_1 = require("../../spec-configuration/containerFeaturesOCI");
const cliHost_1 = require("../../spec-common/cliHost");
const commonUtils_1 = require("../../spec-common/commonUtils");
const packageCommandImpl_2 = require("../collectionCommonUtils/packageCommandImpl");
const targetPositionalDescription = `
Package and publish features at provided [target] (default is cwd), where [target] is either:
   1. A path to the src folder of the collection with [1..n] features.
   2. A path to a single feature that contains a devcontainer-feature.json.
`;
function featuresPublishOptions(y) {
    return y
        .options({
        'registry': { type: 'string', alias: 'r', default: 'ghcr.io', description: 'Name of the OCI registry.' },
        'namespace': { type: 'string', alias: 'n', require: true, description: 'Unique indentifier for the collection of features. Example: <owner>/<repo>' },
        'log-level': { choices: ['info', 'debug', 'trace'], default: 'info', description: 'Log level.' }
    })
        .positional('target', { type: 'string', default: '.', description: targetPositionalDescription })
        .check(_argv => {
        return true;
    });
}
exports.featuresPublishOptions = featuresPublishOptions;
function featuresPublishHandler(args) {
    (async () => await featuresPublish(args))().catch(console.error);
}
exports.featuresPublishHandler = featuresPublishHandler;
async function featuresPublish({ 'target': targetFolder, 'log-level': inputLogLevel, 'registry': registry, 'namespace': namespace }) {
    const disposables = [];
    const dispose = async () => {
        await Promise.all(disposables.map(d => d()));
    };
    const pkg = (0, product_1.getPackageConfig)();
    const cwd = process.cwd();
    const cliHost = await (0, cliHost_1.getCLIHost)(cwd, commonUtils_1.loadNativeModule);
    const output = (0, devContainers_1.createLog)({
        logLevel: (0, log_1.mapLogLevel)(inputLogLevel),
        logFormat: 'text',
        log: (str) => process.stdout.write(str),
        terminalDimensions: undefined,
    }, pkg, new Date(), disposables);
    // Package features
    const outputDir = path_1.default.join(os.tmpdir(), '/features-output');
    const packageArgs = {
        cliHost,
        targetFolder,
        outputDir,
        output,
        disposables,
        forceCleanOutputDir: true,
    };
    const metadata = await (0, packageCommandImpl_1.doFeaturesPackageCommand)(packageArgs);
    if (!metadata) {
        output.write(`(!) ERR: Failed to fetch ${packageCommandImpl_2.OCICollectionFileName}`, log_1.LogLevel.Error);
        process.exit(1);
    }
    for (const f of metadata.features) {
        output.write(`Processing feature: ${f.id}...`, log_1.LogLevel.Info);
        if (!f.version) {
            output.write(`(!) WARNING: Version does not exist, skipping ${f.id}...`, log_1.LogLevel.Warning);
            continue;
        }
        const resource = `${registry}/${namespace}/${f.id}`;
        const featureRef = (0, containerFeaturesOCI_1.getFeatureRef)(output, resource);
        await (0, publishCommandImpl_1.doFeaturesPublishCommand)(f.version, featureRef, outputDir, output);
    }
    const featureCollectionRef = {
        registry: registry,
        path: namespace,
        version: 'latest'
    };
    await (0, publishCommandImpl_1.doFeaturesPublishMetadata)(featureCollectionRef, outputDir, output);
    // Cleanup
    await (0, pfs_1.rmLocal)(outputDir, { recursive: true, force: true });
    await dispose();
    process.exit();
}
//# sourceMappingURL=publish.js.map