"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.doFeaturesPublishMetadata = exports.doFeaturesPublishCommand = exports.getSermanticVersions = void 0;
const path_1 = __importDefault(require("path"));
const semver = __importStar(require("semver"));
const containerFeaturesOCI_1 = require("../../spec-configuration/containerFeaturesOCI");
const containerFeaturesOCIPush_1 = require("../../spec-configuration/containerFeaturesOCIPush");
const log_1 = require("../../spec-utils/log");
const packageCommandImpl_1 = require("../collectionCommonUtils/packageCommandImpl");
const packageCommandImpl_2 = require("./packageCommandImpl");
let semanticVersions = [];
function updateSemanticVersionsList(publishedVersions, version, range, publishVersion) {
    // Reference: https://github.com/npm/node-semver#ranges-1
    const publishedMaxVersion = semver.maxSatisfying(publishedVersions, range);
    if (publishedMaxVersion === null || semver.compare(version, publishedMaxVersion) === 1) {
        semanticVersions.push(publishVersion);
    }
    return;
}
function getSermanticVersions(version, publishedVersions, output) {
    if (publishedVersions.includes(version)) {
        output.write(`(!) WARNING: Version ${version} already exists, skipping ${version}...`, log_1.LogLevel.Warning);
        return undefined;
    }
    const parsedVersion = semver.parse(version);
    if (!parsedVersion) {
        output.write(`(!) ERR: Version ${version} is not a valid semantic version, skipping ${version}...`, log_1.LogLevel.Error);
        process.exit(1);
    }
    semanticVersions = [];
    // Adds semantic versions depending upon the existings (published) versions
    // eg. 1.2.3 --> [1, 1.2, 1.2.3, latest]
    updateSemanticVersionsList(publishedVersions, version, `${parsedVersion.major}.x.x`, `${parsedVersion.major}`);
    updateSemanticVersionsList(publishedVersions, version, `${parsedVersion.major}.${parsedVersion.minor}.x`, `${parsedVersion.major}.${parsedVersion.minor}`);
    semanticVersions.push(version);
    updateSemanticVersionsList(publishedVersions, version, `x.x.x`, 'latest');
    return semanticVersions;
}
exports.getSermanticVersions = getSermanticVersions;
async function doFeaturesPublishCommand(version, featureRef, outputDir, output) {
    output.write(`Fetching published versions...`, log_1.LogLevel.Info);
    const publishedVersions = await (0, containerFeaturesOCI_1.getPublishedVersions)(featureRef, output);
    if (!publishedVersions) {
        process.exit(1);
    }
    const semanticVersions = getSermanticVersions(version, publishedVersions, output);
    if (!!semanticVersions) {
        output.write(`Publishing versions: ${semanticVersions.toString()}...`, log_1.LogLevel.Info);
        const pathToTgz = path_1.default.join(outputDir, (0, packageCommandImpl_2.getFeatureArchiveName)(featureRef.id));
        if (!await (0, containerFeaturesOCIPush_1.pushOCIFeature)(output, featureRef, pathToTgz, semanticVersions)) {
            output.write(`(!) ERR: Failed to publish feature: ${featureRef.resource}`, log_1.LogLevel.Error);
            process.exit(1);
        }
        output.write(`Published feature: ${featureRef.id}...`, log_1.LogLevel.Info);
    }
}
exports.doFeaturesPublishCommand = doFeaturesPublishCommand;
async function doFeaturesPublishMetadata(featureCollectionRef, outputDir, output) {
    // Publishing Feature Collection Metadata
    output.write('Publishing collection metadata...', log_1.LogLevel.Info);
    const pathToFeatureCollectionFile = path_1.default.join(outputDir, packageCommandImpl_1.OCICollectionFileName);
    if (!await (0, containerFeaturesOCIPush_1.pushFeatureCollectionMetadata)(output, featureCollectionRef, pathToFeatureCollectionFile)) {
        output.write(`(!) ERR: Failed to publish collection metadata: ${packageCommandImpl_1.OCICollectionFileName}`, log_1.LogLevel.Error);
        process.exit(1);
    }
    output.write('Published collection metadata...', log_1.LogLevel.Info);
}
exports.doFeaturesPublishMetadata = doFeaturesPublishMetadata;
//# sourceMappingURL=publishCommandImpl.js.map