"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.featuresTestHandler = exports.featuresTestOptions = void 0;
const cliHost_1 = require("../../spec-common/cliHost");
const commonUtils_1 = require("../../spec-common/commonUtils");
const log_1 = require("../../spec-utils/log");
const product_1 = require("../../spec-utils/product");
const testCommandImpl_1 = require("./testCommandImpl");
// -- 'features test' command
function featuresTestOptions(y) {
    return y
        .options({
        'features': { type: 'array', alias: 'f', describe: 'Feature(s) to test as space-separated parameters. Omit to run all tests.  Cannot be combined with \'--global-scenarios-only\'.', },
        'global-scenarios-only': { type: 'boolean', default: false, description: 'Run only scenario tests under \'tests/_global\' .  Cannot be combined with \'-f\'.' },
        'skip-scenarios': { type: 'boolean', default: false, description: 'Skip all \'scenario\' style tests.  Cannot be combined with \'--global--scenarios-only\'.' },
        'skip-autogenerated': { type: 'boolean', default: false, description: 'Skip all \'autogenerated\' style tests.' },
        'base-image': { type: 'string', alias: 'i', default: 'ubuntu:focal', description: 'Base Image' },
        'remote-user': { type: 'string', alias: 'u', default: 'root', describe: 'Remote user', },
        'log-level': { choices: ['info', 'debug', 'trace'], default: 'info', description: 'Log level.' },
        'quiet': { type: 'boolean', alias: 'q', default: false, description: 'Quiets output' },
    })
        .positional('target', { type: 'string', default: '.', description: 'Path to folder containing \'src\' and \'test\' sub-folders.' })
        .check(argv => {
        if (argv['global-scenarios-only'] && argv['features']) {
            throw new Error('Cannot combine --global-scenarios-only and --features');
        }
        if (argv['skip-scenarios'] && argv['global-scenarios-only']) {
            throw new Error('Cannot combine --skip-scenarios and --global-scenarios-only');
        }
        return true;
    });
}
exports.featuresTestOptions = featuresTestOptions;
function featuresTestHandler(args) {
    (async () => await featuresTest(args))().catch(console.error);
}
exports.featuresTestHandler = featuresTestHandler;
async function featuresTest({ 'base-image': baseImage, 'target': collectionFolder, features, 'global-scenarios-only': globalScenariosOnly, 'skip-scenarios': skipScenarios, 'skip-autogenerated': skipAutogenerated, 'remote-user': remoteUser, quiet, 'log-level': inputLogLevel, }) {
    const disposables = [];
    const dispose = async () => {
        await Promise.all(disposables.map(d => d()));
    };
    const cwd = process.cwd();
    const cliHost = await (0, cliHost_1.getCLIHost)(cwd, commonUtils_1.loadNativeModule);
    const pkg = (0, product_1.getPackageConfig)();
    const logLevel = (0, log_1.mapLogLevel)(inputLogLevel);
    const args = {
        baseImage,
        cliHost,
        logLevel,
        quiet,
        pkg,
        collectionFolder: cliHost.path.resolve(collectionFolder),
        features: features ? (Array.isArray(features) ? features : [features]) : undefined,
        globalScenariosOnly,
        skipScenarios,
        skipAutogenerated,
        remoteUser,
        disposables
    };
    const exitCode = await (0, testCommandImpl_1.doFeaturesTestCommand)(args);
    await dispose();
    process.exit(exitCode);
}
// -- End: 'features test' command
//# sourceMappingURL=test.js.map