import { DevContainerConfig, DevContainerConfigCommand, DevContainerFromDockerComposeConfig, DevContainerFromDockerfileConfig, DevContainerFromImageConfig, HostRequirements, PortAttributes, UserEnvProbe } from '../spec-configuration/configuration';
import { Feature, FeaturesConfig, Mount } from '../spec-configuration/containerFeaturesConfiguration';
import { ContainerDetails, DockerCLIParameters, ImageDetails } from '../spec-shutdown/dockerUtils';
import { Log } from '../spec-utils/log';
import { SubstituteConfig, SubstitutedConfig, DockerResolverParameters } from './utils';
export interface ImageMetadataEntry {
    id?: string;
    init?: boolean;
    privileged?: boolean;
    capAdd?: string[];
    securityOpt?: string[];
    entrypoint?: string;
    mounts?: (Mount | string)[];
    customizations?: Record<string, any>;
    onCreateCommand?: string | string[];
    updateContentCommand?: string | string[];
    postCreateCommand?: string | string[];
    postStartCommand?: string | string[];
    postAttachCommand?: string | string[];
    waitFor?: DevContainerConfigCommand;
    remoteUser?: string;
    containerUser?: string;
    userEnvProbe?: UserEnvProbe;
    remoteEnv?: Record<string, string | null>;
    containerEnv?: Record<string, string>;
    overrideCommand?: boolean;
    portsAttributes?: Record<string, PortAttributes>;
    otherPortsAttributes?: PortAttributes;
    forwardPorts?: (number | string)[];
    shutdownAction?: 'none' | 'stopContainer' | 'stopCompose';
    updateRemoteUserUID?: boolean;
    hostRequirements?: HostRequirements;
}
export declare type MergedDevContainerConfig = MergedConfig<DevContainerFromImageConfig> | MergedConfig<DevContainerFromDockerfileConfig> | MergedConfig<DevContainerFromDockerComposeConfig>;
declare type MergedConfig<T extends DevContainerConfig> = Omit<T, typeof replaceProperties[number]> & UpdatedConfigProperties;
declare const replaceProperties: readonly ["customizations", "entrypoint", "onCreateCommand", "updateContentCommand", "postCreateCommand", "postStartCommand", "postAttachCommand", "shutdownAction"];
interface UpdatedConfigProperties {
    customizations?: Record<string, any[]>;
    entrypoints?: string[];
    onCreateCommands?: (string | string[])[];
    updateContentCommands?: (string | string[])[];
    postCreateCommands?: (string | string[])[];
    postStartCommands?: (string | string[])[];
    postAttachCommands?: (string | string[])[];
    shutdownAction?: 'none' | 'stopContainer' | 'stopCompose';
}
export declare function mergeConfiguration(config: DevContainerConfig, imageMetadata: ImageMetadataEntry[]): MergedDevContainerConfig;
export declare function getDevcontainerMetadata(baseImageMetadata: SubstitutedConfig<ImageMetadataEntry[]>, devContainerConfig: SubstitutedConfig<DevContainerConfig>, featuresConfig: FeaturesConfig | Feature[]): SubstitutedConfig<ImageMetadataEntry[]>;
export interface ImageBuildInfo {
    user: string;
    metadata: SubstitutedConfig<ImageMetadataEntry[]>;
}
export declare function getImageBuildInfo(params: DockerResolverParameters | DockerCLIParameters, configWithRaw: SubstitutedConfig<DevContainerConfig>, experimentalImageMetadata: boolean): Promise<ImageBuildInfo>;
export declare function getImageBuildInfoFromImage(params: DockerResolverParameters | DockerCLIParameters, imageName: string, substitute: SubstituteConfig, experimentalImageMetadata: boolean): Promise<ImageBuildInfo & {
    imageDetails: ImageDetails;
}>;
export declare function getImageBuildInfoFromDockerfile(params: DockerResolverParameters | DockerCLIParameters, dockerfile: string, dockerBuildArgs: Record<string, string>, targetStage: string | undefined, substitute: SubstituteConfig, experimentalImageMetadata: boolean): Promise<ImageBuildInfo>;
export declare function internalGetImageBuildInfoFromDockerfile(inspectDockerImage: (imageName: string) => Promise<ImageDetails>, dockerfile: string, dockerBuildArgs: Record<string, string>, targetStage: string | undefined, substitute: SubstituteConfig, experimentalImageMetadata: boolean, output: Log): Promise<ImageBuildInfo>;
export declare const imageMetadataLabel = "devcontainer.metadata";
export declare function getImageMetadataFromContainer(containerDetails: ContainerDetails, devContainerConfig: SubstitutedConfig<DevContainerConfig>, featuresConfig: FeaturesConfig | Feature[], experimentalImageMetadata: boolean, output: Log): SubstitutedConfig<ImageMetadataEntry[]>;
export declare function getImageMetadata(imageDetails: ImageDetails, substitute: SubstituteConfig, experimentalImageMetadata: boolean, output: Log): SubstitutedConfig<ImageMetadataEntry[]>;
export declare function getDevcontainerMetadataLabel(baseImageMetadata: SubstitutedConfig<ImageMetadataEntry[]>, devContainerConfig: SubstitutedConfig<DevContainerConfig>, featuresConfig: FeaturesConfig | Feature[], experimentalImageMetadata: boolean): string;
export {};
