"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.templatesPackageHandler = exports.templatesPackageOptions = void 0;
const cliHost_1 = require("../../spec-common/cliHost");
const commonUtils_1 = require("../../spec-common/commonUtils");
const log_1 = require("../../spec-utils/log");
const product_1 = require("../../spec-utils/product");
const devContainers_1 = require("../devContainers");
const package_1 = require("../collectionCommonUtils/package");
const packageCommandImpl_1 = require("./packageCommandImpl");
function templatesPackageOptions(y) {
    return (0, package_1.PackageOptions)(y, 'template');
}
exports.templatesPackageOptions = templatesPackageOptions;
function templatesPackageHandler(args) {
    (async () => await templatesPackage(args))().catch(console.error);
}
exports.templatesPackageHandler = templatesPackageHandler;
async function templatesPackage({ 'target': targetFolder, 'log-level': inputLogLevel, 'output-folder': outputDir, 'force-clean-output-folder': forceCleanOutputDir, }) {
    const disposables = [];
    const dispose = async () => {
        await Promise.all(disposables.map(d => d()));
    };
    const pkg = (0, product_1.getPackageConfig)();
    const cwd = process.cwd();
    const cliHost = await (0, cliHost_1.getCLIHost)(cwd, commonUtils_1.loadNativeModule);
    const output = (0, devContainers_1.createLog)({
        logLevel: (0, log_1.mapLogLevel)(inputLogLevel),
        logFormat: 'text',
        log: (str) => process.stdout.write(str),
        terminalDimensions: undefined,
    }, pkg, new Date(), disposables);
    const args = {
        cliHost,
        targetFolder,
        outputDir,
        output,
        disposables,
        forceCleanOutputDir: forceCleanOutputDir
    };
    const exitCode = !!(await (0, packageCommandImpl_1.doTemplatesPackageCommand)(args)) ? 0 : 1;
    await dispose();
    process.exit(exitCode);
}
//# sourceMappingURL=package.js.map