"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.packageTemplateCollection = exports.packageSingleTemplate = exports.doTemplatesPackageCommand = void 0;
const path_1 = __importDefault(require("path"));
const packageCommandImpl_1 = require("../collectionCommonUtils/packageCommandImpl");
const log_1 = require("../../spec-utils/log");
const pfs_1 = require("../../spec-utils/pfs");
const collectionType = 'template';
async function doTemplatesPackageCommand(args) {
    args = await (0, packageCommandImpl_1.prepPackageCommand)(args, collectionType);
    const { output, outputDir } = args;
    const isSingleTemplate = args.isSingle;
    // For each template, package each template and write to 'outputDir/{f}.tgz'
    // Returns an array of template metadata from each processed template
    let metadataOutput = [];
    if (isSingleTemplate) {
        // Package individual templates
        output.write('Packaging single template...', log_1.LogLevel.Info);
        metadataOutput = await packageSingleTemplate(args);
    }
    else {
        output.write('Packaging template collection...', log_1.LogLevel.Info);
        metadataOutput = await packageTemplateCollection(args);
    }
    if (!metadataOutput) {
        output.write('Failed to package templates', log_1.LogLevel.Error);
        return undefined;
    }
    const collection = {
        sourceInformation: {
            source: 'devcontainer-cli',
        },
        templates: metadataOutput,
    };
    // Write the metadata to a file
    const metadataOutputPath = path_1.default.join(outputDir, packageCommandImpl_1.OCICollectionFileName);
    await (0, pfs_1.writeLocalFile)(metadataOutputPath, JSON.stringify(collection, null, 4));
    return collection;
}
exports.doTemplatesPackageCommand = doTemplatesPackageCommand;
async function packageSingleTemplate(args) {
    return await (0, packageCommandImpl_1.packageSingleFeatureOrTemplate)(args, collectionType);
}
exports.packageSingleTemplate = packageSingleTemplate;
async function packageTemplateCollection(args) {
    return await (0, packageCommandImpl_1.packageCollection)(args, collectionType);
}
exports.packageTemplateCollection = packageTemplateCollection;
//# sourceMappingURL=packageCommandImpl.js.map